import itertools

import pytest

from diofant import (ITE, And, EmptySet, Equality, Equivalent, Implies,
                     Integer, Interval, Nand, Nor, Not, Or, S, Unequality,
                     Union, Xor, false, oo, simplify_logic, sqrt, to_cnf,
                     to_dnf, to_nnf, true)
from diofant.abc import a, b, c, d, w, x, y, z
from diofant.logic.boolalg import (BooleanAtom, BooleanFunction, _POSform,
                                   _SOPform, is_cnf, is_dnf, is_literal,
                                   is_nnf, to_int_repr)


__all__ = ()


def test_overloading():
    assert a & b == And(a, b)
    assert a | b == Or(a, b)
    assert a >> b == Implies(a, b)
    assert ~a == Not(a)
    assert a ^ b == Xor(a, b)


def test_And():
    assert And() is true
    assert And(a) == a
    assert And(True) is true
    assert And(False) is false
    assert And(True, True) is true
    assert And(True, False) is false
    assert And(True, False, evaluate=False) is not false
    assert And(False, False) is false
    assert And(True, a) == a
    assert And(False, a) is false
    assert And(True, True, True) is true
    assert And(True, True, a) == a
    assert And(True, False, a) is false
    assert And(2, a) == a
    assert And(2, 3) is true
    assert And(a < 1, a >= 1) is false

    e = a > 1
    assert And(e, e.canonical) == e.canonical

    g, l, ge, le = a > b, b < a, a >= b, b <= a
    assert And(g, l, ge, le) == And(l, le)


def test_Or():
    assert Or() is false
    assert Or(a) == a
    assert Or(True) is true
    assert Or(False) is false
    assert Or(True, True) is true
    assert Or(True, False) is true
    assert Or(False, False) is false
    assert Or(True, a) is true
    assert Or(False, a) == a
    assert Or(True, False, False) is true
    assert Or(True, False, a) is true
    assert Or(False, False, a) == a
    assert Or(2, a) is true
    assert Or(a < 1, a >= 1) is true

    e = a > 1
    assert Or(e, e.canonical) == e

    g, l, ge, le = a > b, b < a, a >= b, b <= a
    assert Or(g, l, ge, le) == Or(g, ge)


def test_Xor():
    assert Xor() is false
    assert Xor(a) == a
    assert Xor(a, a) is false
    assert Xor(True, a, a) is true
    assert Xor(a, a, a, a, a) == a
    assert Xor(True, False, False, a, b) == ~Xor(a, b)
    assert Xor(True) is true
    assert Xor(False) is false
    assert Xor(True, True) is false
    assert Xor(True, False) is true
    assert Xor(False, False) is false
    assert Xor(True, a) == ~a
    assert Xor(False, a) == a
    assert Xor(True, False, False) is true
    assert Xor(True, False, a) == ~a
    assert Xor(False, False, a) == a
    assert isinstance(Xor(a, b), Xor)
    assert Xor(a, b, Xor(c, d)) == Xor(a, b, c, d)
    assert Xor(a, b, Xor(b, c)) == Xor(a, c)
    assert Xor(a < 1, a >= 1, b) == Xor(0, 1, b) == Xor(1, 0, b)

    e = a > 1
    assert Xor(e, e.canonical) == Xor(0, 0) == Xor(1, 1)

    e = Integer(1) < a
    assert e != e.canonical
    assert Xor(e, e.canonical) is false

    assert Xor(a > 1, b > c) == Xor(a > 1, b > c, evaluate=False)


def test_Not():
    pytest.raises(TypeError, lambda: Not(True, False))
    assert Not(True) is false
    assert Not(False) is true
    assert Not(0) is true
    assert Not(1) is false
    assert Not(2) is false
    assert Not(Unequality(a, b)) == Equality(a, b)


def test_Nand():
    assert Nand() is false
    assert Nand(a) == ~a
    assert Nand(True) is false
    assert Nand(False) is true
    assert Nand(True, True) is false
    assert Nand(True, False) is true
    assert Nand(False, False) is true
    assert Nand(True, a) == ~a
    assert Nand(False, a) is true
    assert Nand(True, True, True) is false
    assert Nand(True, True, a) == ~a
    assert Nand(True, False, a) is true


def test_Nor():
    assert Nor() is true
    assert Nor(a) == ~a
    assert Nor(True) is false
    assert Nor(False) is true
    assert Nor(True, True) is false
    assert Nor(True, False) is false
    assert Nor(False, False) is true
    assert Nor(True, a) is false
    assert Nor(False, a) == ~a
    assert Nor(True, True, True) is false
    assert Nor(True, True, a) is false
    assert Nor(True, False, a) is false


def test_Implies():
    pytest.raises(ValueError, lambda: Implies(a, b, c))

    assert Implies(True, True) is true
    assert Implies(True, False) is false
    assert Implies(False, True) is true
    assert Implies(False, False) is true
    assert Implies(0, a) is true
    assert Implies(1, 1) is true
    assert Implies(1, 0) is false
    assert a >> b == b << a
    assert (a < 1) >> (a >= 1) == (a >= 1)
    assert (a < 1) >> (1 > a) is true
    assert (a < 1) >> (Integer(1) > a) is true
    assert a >> a is true
    assert ((a < 1) >> (b >= 1)) == Implies(a < 1, b >= 1, evaluate=False)


def test_Equivalent():
    assert Equivalent(a, b) == Equivalent(b, a) == Equivalent(a, b, a)
    assert Equivalent() is true
    assert Equivalent(a, a) == Equivalent(a) is true
    assert Equivalent(True, True) == Equivalent(False, False) is true
    assert Equivalent(True, False) == Equivalent(False, True) is false
    assert Equivalent(a, True) == a
    assert Equivalent(a, False) == ~a
    assert Equivalent(a, b, True) == a & b
    assert Equivalent(a, b, False) == ~a & ~b
    assert Equivalent(1, a) == a
    assert Equivalent(0, a) == Not(a)
    assert Equivalent(a, Equivalent(b, c)) != Equivalent(Equivalent(a, b), c)
    assert Equivalent(a < 1, a >= 1) is false
    assert Equivalent(a < 1, a >= 1, 0) is false
    assert Equivalent(a < 1, a >= 1, 1) is false
    assert Equivalent(a < 1, 1 > a) == Equivalent(1, 1) == Equivalent(0, 0)
    assert Equivalent(a < 1, Integer(1) > a) == Equivalent(1, 1) == Equivalent(0, 0)
    assert Equivalent(a < 1, b >= 1) == Equivalent(b >= 1, a < 1, evaluate=False)


def test_equals():
    assert (~(a | b)).equals(~a & ~b) is True
    assert Equivalent(a, b).equals((a >> b) & (b >> a)) is True
    assert ((a | ~b) & (~a | b)).equals((~a & ~b) | (a & b)) is True
    assert (a >> b).equals(~a >> ~b) is False
    assert (a >> (b >> a)).equals(a >> (c >> a)) is False
    pytest.raises(NotImplementedError, lambda: (a & (a < b)).equals(a & (b > a)))


def test_simplification():
    set1 = [[0, 0, 1], [0, 1, 1], [1, 0, 0], [1, 1, 0]]
    set2 = [[0, 0, 0], [0, 1, 0], [1, 0, 1], [1, 1, 1]]
    assert _SOPform([x, y, z], set1) == (~x & z) | (~z & x)
    assert ~_SOPform([x, y, z], set2) == ~((~x & ~z) | (x & z))
    assert _POSform([x, y, z], set1 + set2) is true
    assert _SOPform([x, y, z], set1 + set2) is true
    assert _SOPform([w, x, y, z], set1 + set2) is true

    minterms = [[0, 0, 0, 1], [0, 0, 1, 1], [0, 1, 1, 1], [1, 0, 1, 1],
                [1, 1, 1, 1]]
    dontcares = [[0, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 1]]
    assert _SOPform([w, x, y, z], minterms, dontcares) == (~w & z) | (y & z)
    assert _POSform([w, x, y, z], minterms, dontcares) == (~w | y) & z

    # test simplification
    assert (a & (b | c)).simplify() == a & (b | c)
    assert ((a & b) | (a & c)).simplify() == a & (b | c)
    assert (a >> b).simplify() == ~a | b
    assert to_dnf(Equivalent(a, b), simplify=True) == (a & b) | (~a & ~b)
    assert (Equality(a, 2) & c).simplify() == Equality(a, 2) & c
    assert (Equality(a, 2) & a).simplify() == Equality(a, 2) & a
    assert (Equality(a, b) & c).simplify() == Equality(a, b) & c
    assert ((Equality(a, 3) & b) | (Equality(a, 3) & c)).simplify() == Equality(a, 3) & (b | c)

    e = a & (x**2 - x)
    assert e.simplify() == a & x*(x - 1)
    assert simplify_logic(e, deep=False) == e

    pytest.raises(ValueError, lambda: simplify_logic(a & (b | c), form='spam'))

    e = x & y ^ z | (z ^ x)
    res = [(x & ~z) | (z & ~x) | (z & ~y), (x & ~y) | (x & ~z) | (z & ~x)]
    assert to_dnf(e, simplify=True) in res
    assert _SOPform([z, y, x], [[0, 0, 1], [0, 1, 1],
                                [1, 0, 0], [1, 0, 1], [1, 1, 0]]) == res[1]

    # check input
    ans = _SOPform([x, y], [[1, 0]])
    assert _SOPform([x, y], [[1, 0]]) == ans
    assert _POSform([x, y], [[1, 0]]) == ans

    pytest.raises(ValueError, lambda: _SOPform([x], [[1]], [[1]]))
    assert _SOPform([x], [[1]], [[0]]) is true
    assert _SOPform([x], [[0]], [[1]]) is true
    assert _SOPform([x], [], []) is false

    pytest.raises(ValueError, lambda: _POSform([x], [[1]], [[1]]))
    assert _POSform([x], [[1]], [[0]]) is true
    assert _POSform([x], [[0]], [[1]]) is true
    assert _POSform([x], [], []) is false

    # check working of simplify
    assert ((a & b) | (a & c)).simplify() == a & (b | c)
    assert (x & ~x).simplify() is false
    assert (x | ~x).simplify() is true


def test_bool_symbol():
    assert And(a, True) == a
    assert And(a, True, True) == a
    assert And(a, False) is false
    assert And(a, True, False) is false
    assert Or(a, True) is true
    assert Or(a, False) == a


def test_is_boolean():
    assert true.is_Boolean
    assert (a & b).is_Boolean
    assert (a | b).is_Boolean
    assert (~a).is_Boolean
    assert (a ^ b).is_Boolean


def test_subs():
    assert (a & b).subs({a: True}) == b
    assert (a & b).subs({a: False}) is false
    assert (a & b).subs({b: True}) == a
    assert (a & b).subs({b: False}) is false
    assert (a & b).subs({a: True, b: True}) is true
    assert (a | b).subs({a: True}) is true
    assert (a | b).subs({a: False}) == b
    assert (a | b).subs({b: True}) is true
    assert (a | b).subs({b: False}) == a
    assert (a | b).subs({a: True, b: True}) is true


def test_logic_commutatity():
    assert a & b == b & a
    assert a | b == b | a


def test_logic_associativity():
    assert (a & b) & c == a & (b & c)
    assert (a | b) | c == a | (b | c)


def test_double_negation():
    assert ~(~a) == a


def test_to_nnf():
    assert to_nnf(true) is true
    assert to_nnf(false) is false
    assert to_nnf(a) == a
    assert to_nnf(~a) == ~a

    class Foo(BooleanFunction):
        pass

    pytest.raises(ValueError, lambda: to_nnf(~Foo(a)))

    assert to_nnf(a | ~a | b) is true
    assert to_nnf(a & ~a & b) is false
    assert to_nnf(a >> b) == ~a | b
    assert to_nnf(Implies(a, b, evaluate=False)) == ~a | b
    assert to_nnf(a >> (c >> ~b)) == (~b | ~c) | ~a
    assert to_nnf(Equivalent(a, b)) == (a | ~b) & (b | ~a)
    assert to_nnf(Equivalent(a, b, c)) == (~a | b) & (~b | c) & (~c | a)
    assert to_nnf(Equivalent(a, b, c, d)) == (~a | b) & (~b | c) & (~c | d) & (~d | a)
    assert to_nnf(a ^ b ^ c) == (a | b | c) & (~a | ~b | c) & (a | ~b | ~c) & (~a | b | ~c)
    assert to_nnf(ITE(a, b, c)) == (~a | b) & (a | c)
    assert to_nnf(~(a | b | c)) == ~a & ~b & ~c
    assert to_nnf(~(a & b & c)) == ~a | ~b | ~c
    assert to_nnf(~(a >> b)) == a & ~b
    assert to_nnf(~(Equivalent(a, b, c))) == (a | b | c) & (~a | ~b | ~c)
    assert to_nnf(~(a ^ b ^ c)) == (~a | b | c) & (a | ~b | c) & (a | b | ~c) & (~a | ~b | ~c)
    assert to_nnf(~(ITE(a, b, c))) == (~a | ~b) & (a | ~c)
    assert to_nnf((a >> b) ^ (b >> a)) == (a & ~b) | (~a & b)
    assert to_nnf((a >> b) ^ (b >> a), False) == (~a | ~b | a | b) & ((a & ~b) | (~a & b))


def test_to_cnf():
    assert to_cnf(~(b | c)) == ~b & ~c
    assert to_cnf((a & b) | c) == (a | c) & (b | c)
    assert to_cnf(a >> b) == ~a | b
    assert to_cnf(a >> (b & c)) == (~a | b) & (~a | c)
    assert to_cnf(a & (b | c) | ~a & (b | c), True) == b | c
    assert to_cnf(a | (~b & ~c)) == (a | ~b) & (a | ~c)

    assert to_cnf(Equivalent(a, b)) == (a | ~b) & (b | ~a)
    assert to_cnf(Equivalent(a, b & c)) == (~a | b) & (~a | c) & (~b | ~c | a)
    assert to_cnf(Equivalent(a, b | c), True) == (~b | a) & (~c | a) & (b | c | ~a)
    assert to_cnf(~(a | b) | c) == (~a | c) & (~b | c)
    assert to_cnf((a & b) | c) == (a | c) & (b | c)


def test_to_dnf():
    assert to_dnf(true) == true
    assert to_dnf(~b & ~c) == ~b & ~c
    assert to_dnf(~(b | c)) == ~b & ~c
    assert to_dnf(a & (b | c)) == (a & b) | (a & c)
    assert to_dnf((~a | b) & c) == (b & c) | (c & ~a)
    assert to_dnf(a >> b) == ~a | b
    assert to_dnf(a >> (b & c)) == (~a) | (b & c)

    assert to_dnf(Equivalent(a, b), True) == (a & b) | (~a & ~b)
    assert to_dnf(Equivalent(a, b & c), True) == ((a & b & c) | (~a & ~b) | (~a & ~c))


def test_to_int_repr():
    assert to_int_repr([x | y, z | x], [x, y, z]) == [{1, 2}, {1, 3}]
    assert to_int_repr([x | y, z | ~x], [x, y, z]) == [{1, 2}, {-1, 3}]


def test_is_nnf():
    assert is_nnf(true) is True
    assert is_nnf(a) is True
    assert is_nnf(~a) is True
    assert is_nnf(a & b) is True
    assert is_nnf((a & b) | (~a & a) | (~b & b) | (~a & ~b), False) is True
    assert is_nnf((a | b) & (~a | ~b)) is True
    assert is_nnf(~(a | b)) is False
    assert is_nnf(a ^ b) is False
    assert is_nnf((a & b) | (~a & a) | (~b & b) | (~a & ~b), True) is False


def test_is_cnf():
    assert is_cnf(x) is True
    assert is_cnf(x | y | z) is True
    assert is_cnf(x & y & z) is True
    assert is_cnf((x | y) & z) is True
    assert is_cnf((x & y) | z) is False
    assert is_cnf(x & y & (z | ~(x ^ y))) is False
    assert is_cnf(x & y & (z | (x ^ y))) is False


def test_is_dnf():
    assert is_dnf(x) is True
    assert is_dnf(x | y | z) is True
    assert is_dnf(x & y & z) is True
    assert is_dnf((x & y) | z) is True
    assert is_dnf((x | y) & z) is False


def test_ITE():
    pytest.raises(ValueError, lambda: ITE(a, b))

    assert ITE(True, False, True) is false
    assert ITE(True, True, False) is true
    assert ITE(False, True, False) is false
    assert ITE(False, False, True) is true
    assert isinstance(ITE(a, b, c), ITE)

    assert ITE(True, b, c) == b
    assert ITE(False, b, c) == c

    assert ITE(a, b, b) == b

    assert ITE(c, False, True) == ~c
    assert ITE(c, True, False) == c

    assert ITE(a > 0, a**2, a).diff(a) == ITE(a > 0, 2*a, 1)


def test_is_literal():
    assert is_literal(True) is True
    assert is_literal(False) is True
    assert is_literal(a) is True
    assert is_literal(~a) is True
    assert is_literal(a | b) is False


def test_logic_operators():
    assert (True & a) == (a & True) == a
    assert (False & a) == (a & False) == false
    assert (True | a) == (a | True) == true
    assert (False | a) == (a | False) == a
    assert (True >> a) == (a << True) == a
    assert (False >> a) == (a << False) == true
    assert (a >> True) == (True << a) == true
    assert (a >> False) == (False << a) == ~a
    assert (True ^ a) == (a ^ True) == ~a
    assert (False ^ a) == (a ^ False) == a


def test_true_false():
    # pylint: disable=singleton-comparison,comparison-with-itself,unneeded-not
    assert true is true
    assert false is false
    assert true is not True
    assert false is not False
    assert true
    assert not false
    assert true == True  # noqa: E712
    assert false == False  # noqa: E712
    assert not true == False  # noqa: E712
    assert not false == True  # noqa: E712
    assert not true == false

    assert hash(true) == hash(True)
    assert hash(false) == hash(False)
    assert len({true, True}) == len({false, False}) == 1

    assert int(true) == 1
    assert int(false) == 0

    assert isinstance(true, BooleanAtom)
    assert isinstance(false, BooleanAtom)

    assert not isinstance(true, bool)
    assert not isinstance(false, bool)

    assert ~true is false
    assert Not(True) is false
    assert ~false is true
    assert Not(False) is true

    for T, F in itertools.product([True, true], [False, false]):
        assert And(T, F) is false
        assert And(F, T) is false
        assert And(F, F) is false
        assert And(T, T) is true
        assert And(T, x) == x
        assert And(F, x) is false
        if not (T is True and F is False):
            assert T & F is false
            assert F & T is false
        if F is not False:
            assert F & F is false
        if T is not True:
            assert T & T is true

        assert Or(T, F) is true
        assert Or(F, T) is true
        assert Or(F, F) is false
        assert Or(T, T) is true
        assert Or(T, x) is true
        assert Or(F, x) == x
        if not (T is True and F is False):
            assert T | F is true
            assert F | T is true
        if F is not False:
            assert F | F is false
        if T is not True:
            assert T | T is true

        assert Xor(T, F) is true
        assert Xor(F, T) is true
        assert Xor(F, F) is false
        assert Xor(T, T) is false
        assert Xor(T, x) == ~x
        assert Xor(F, x) == x
        if not (T is True and F is False):
            assert T ^ F is true
            assert F ^ T is true
        if F is not False:
            assert F ^ F is false
        if T is not True:
            assert T ^ T is false

        assert Nand(T, F) is true
        assert Nand(F, T) is true
        assert Nand(F, F) is true
        assert Nand(T, T) is false
        assert Nand(T, x) == ~x
        assert Nand(F, x) is true

        assert Nor(T, F) is false
        assert Nor(F, T) is false
        assert Nor(F, F) is true
        assert Nor(T, T) is false
        assert Nor(T, x) is false
        assert Nor(F, x) == ~x

        assert Implies(T, F) is false
        assert Implies(F, T) is true
        assert Implies(F, F) is true
        assert Implies(T, T) is true
        assert Implies(T, x) == x
        assert Implies(F, x) is true
        assert Implies(x, T) is true
        assert Implies(x, F) == ~x
        if not (T is True and F is False):
            assert T >> F is false
            assert F << T is false
            assert F >> T is true
            assert T << F is true
        if F is not False:
            assert F >> F is true
            assert F << F is true
        if T is not True:
            assert T >> T is true
            assert T << T is true

        assert Equivalent(T, F) is false
        assert Equivalent(F, T) is false
        assert Equivalent(F, F) is true
        assert Equivalent(T, T) is true
        assert Equivalent(T, x) == x
        assert Equivalent(F, x) == ~x
        assert Equivalent(x, T) == x
        assert Equivalent(x, F) == ~x

        assert ITE(T, T, T) is true
        assert ITE(T, T, F) is true
        assert ITE(T, F, T) is false
        assert ITE(T, F, F) is false
        assert ITE(F, T, T) is true
        assert ITE(F, T, F) is false
        assert ITE(F, F, T) is true
        assert ITE(F, F, F) is false


def test_bool_as_set():
    assert ((x <= 2) & (x >= -2)).as_set() == Interval(-2, 2)
    assert ((x >= 2) | (x <= -2)).as_set() == (Interval(-oo, -2) + Interval(2, oo, False))
    assert Not(x > 2, evaluate=False).as_set() == Interval(-oo, 2, True)
    assert true.as_set() == S.UniversalSet
    assert false.as_set() == EmptySet()


@pytest.mark.xfail
def test_multivariate_bool_as_set():
    ((x >= 0) & (y >= 0)).as_set()  # == Interval(0, oo)*Interval(0, oo)
    ((x >= 0) | (y >= 0)).as_set()  # == Reals*Reals - Interval(-oo, 0, True, True)*Interval(-oo, 0, True, True)


def test_all_or_nothing():
    args = x >= -oo, x <= oo
    v = And(*args)
    if isinstance(v, And):
        assert len(v.args) == len(args) - args.count(true)
    else:
        assert v
    v = Or(*args)
    if isinstance(v, Or):
        assert len(v.args) == 2
    else:
        assert v


def test_canonical_atoms():
    assert true.canonical == true
    assert false.canonical == false


def test_sympyissue_8777():
    assert ((x > 2) & (x < oo)).as_set() == Interval(2, oo, True, True)
    assert ((x >= 1) & (x < oo)).as_set() == Interval(1, oo, False, True)
    assert (x < oo).as_set() == Interval(-oo, oo, False, True)
    assert (x > -oo).as_set() == Interval(-oo, oo, True)


def test_sympyissue_8975():
    assert (((-oo < x) & (x <= -2)) | ((2 <= x) & (x < oo))).as_set() == Interval(-oo, -2, True) + Interval(2, oo, False, True)


def test_sympyissue_10240():
    assert (~((x > 2) & (x < 3))).as_set() == Union(Interval(-oo, 2, True), Interval(3, oo, False, True))


def test_sympyissue_10641():
    assert str(((x < sqrt(3)) | x).evalf(2)) == 'x | (x < 1.7)'
    assert str(((x < sqrt(3)) & x).evalf(2)) == 'x & (x < 1.7)'


def test_sympyissue_12522():
    assert Equality(1, 1).simplify() is true
    assert true.simplify() is true
    assert false.simplify() is false



"""Gosper's algorithm for hypergeometric summation."""

from ..core import Dummy, Integer, symbols
from ..core.compatibility import is_sequence
from ..polys import parallel_poly_from_expr
from ..simplify import hypersimp
from ..solvers import solve


def gosper_normal(f, g, n):
    r"""Compute the Gosper's normal form of ``f`` and ``g``.

    Given relatively prime univariate polynomials ``f`` and ``g``,
    rewrite their quotient to a normal form defined as follows:

    .. math::
        \frac{f(n)}{g(n)} = Z \cdot \frac{A(n) C(n+1)}{B(n) C(n)}

    where ``Z`` is an arbitrary constant and ``A``, ``B``, ``C`` are
    monic polynomials in ``n`` with the following properties:

    1. `\gcd(A(n), B(n+h)) = 1 \forall h \in \mathbb{N}`
    2. `\gcd(B(n), C(n+1)) = 1`
    3. `\gcd(A(n), C(n)) = 1`

    This normal form, or rational factorization in other words, is a
    crucial step in Gosper's algorithm and in solving of difference
    equations. It can be also used to decide if two hypergeometric
    terms are similar or not.

    This procedure will return a tuple containing elements of this
    factorization in the form ``(Z*A, B, C)``.

    Examples
    ========

    >>> gosper_normal(4*n + 5, 2*(4*n + 1)*(2*n + 3), n)
    (Poly(1/4, n, domain='QQ'), Poly(n + 3/2, n, domain='QQ'),
     Poly(n + 1/4, n, domain='QQ'))

    """
    (C, p, q), _ = parallel_poly_from_expr((1, f, g), n, field=True)

    a, A = p.LC(), p.monic()
    b, B = q.LC(), q.monic()
    Z = a/b

    J = A.dispersionset(B)

    for i in sorted(J):
        d = A.gcd(B.shift(+i))

        A = A.quo(d)
        B = B.quo(d.shift(-i))

        for j in range(1, i + 1):
            C *= d.shift(-j)

    return A*Z, B, C


def gosper_term(f, n):
    r"""Compute Gosper's hypergeometric term for ``f``.

    Suppose ``f`` is a hypergeometric term such that:

    .. math::
        s_n = \sum_{k=0}^{n-1} f_k

    and `f_k` doesn't depend on `n`. Returns a hypergeometric
    term `g_n` such that `g_{n+1} - g_n = f_n`.

    Examples
    ========

    >>> gosper_term((4*n + 1)*factorial(n)/factorial(2*n + 1), n)
    (-n - 1/2)/(n + 1/4)

    """
    r = hypersimp(f, n)

    if r is None:
        return

    p, q = r.as_numer_denom()

    A, B, C = gosper_normal(p, q, n)
    B = B.shift(-1)

    N = Integer(A.degree())
    M = Integer(B.degree())
    K = Integer(C.degree())

    if (N != M) or (A.LC() != B.LC()):
        D = {K - max(N, M)}
    elif not N:
        D = {K - N + 1, Integer(0)}
    else:
        D = {K - N + 1, (B.coeff_monomial(n**(N - 1)) -
                         A.coeff_monomial(n**(N - 1)))/A.LC()}

    for d in set(D):
        if not d.is_Integer or d < 0:
            D.remove(d)

    if not D:
        return

    d = max(D)

    coeffs = symbols(f'c:{d + 1}', cls=Dummy)
    domain = A.domain.inject(*coeffs)

    x = sum(c*n**i for i, c in enumerate(reversed(coeffs)))
    x = x.as_poly(n, domain=domain)
    H = A*x.shift(1) - B*x - C

    solution = solve(H.coeffs(), coeffs)
    if solution:
        solution = solution[0]

    x = x.subs(solution).subs({_: 0 for _ in coeffs})

    if x != 0:
        return B.as_expr()*x.as_expr()/C.as_expr()


def gosper_sum(f, k):
    r"""Gosper's hypergeometric summation algorithm.

    Given a hypergeometric term ``f`` such that:

    .. math ::
        s_n = \sum_{k=0}^{n-1} f_k

    and `f(n)` doesn't depend on `n`, returns `g_{n} - g(0)` where
    `g_{n+1} - g_n = f_n`, or ``None`` if `s_n` can not be expressed
    in closed form as a sum of hypergeometric terms.

    Examples
    ========

    >>> gosper_sum((4*k + 1)*factorial(k)/factorial(2*k + 1), (k, 0, n))
    (-factorial(n) + 2*factorial(2*n + 1))/factorial(2*n + 1)

    References
    ==========

    * :cite:`Petkovsek1997AeqB`

    """
    indefinite = False

    if is_sequence(k):
        k, a, b = k
    else:
        indefinite = True

    g = gosper_term(f, k)

    if g is None:
        return

    if indefinite:
        result = f*g
    else:
        fg = (f*g).cancel()
        result = (f + fg).subs({k: b}) - fg.subs({k: a})

    return result.factor()



"""Square-free decomposition algorithms and related tools."""

from .polyerrors import DomainError


class _SQF:
    """Mixin class to compute square-free decomposition of polynomials."""

    def sqf_list(self, f):
        """
        Return square-free decomposition of a polynomial in ``K[X]``.

        Examples
        ========

        >>> R, x, y = ring('x y', ZZ)

        >>> R.sqf_list(x**5 + 2*x**4*y + x**3*y**2)
        (1, [(x + y, 2), (x, 3)])

        >>> R, x, y = ring('x y', FF(5))

        >>> f = x**5*y**5 + 1

        Note that e.g.:

        >>> f.diff(x)
        0

        This phenomenon doesn't happen in characteristic zero. However we can
        still compute square-free decomposition of ``f``:

        >>> R.sqf_list(f)
        (1, [(x*y + 1, 5)])

        """
        domain = self.domain

        if domain.is_Field:
            coeff, f = f.LC, f.monic()
        else:
            coeff, f = f.primitive()

        if domain.is_FiniteField:
            return coeff, self._gf_sqf_list(f)
        else:
            return coeff, self._rr_yun0_sqf_list(f)

    def _gf_sqf_list(self, f):
        """
        Compute square-free decomposition of the monic ``f`` in ``GF(q)[X]``.

        Notes
        =====

        Uses a modified version of Musser's algorithm for square-free
        decomposition of univariate polynomials over finite fields.

        References
        ==========

        * :cite:`Geddes1992algorithms`, algorithm 8.3
        * :cite:`Musser1971factor`, p.54, algorithm V (modified)

        """
        domain = self.domain

        n, factors, p = 1, [], int(domain.characteristic)
        m = int(domain.order // p)

        while not f.is_ground:
            df = [f.diff(x) for x in self.gens]

            if any(_ for _ in df):
                g = f
                for q in df:
                    g = self.gcd(g, q)
                h, f, i = f // g, g, 1

                while h != 1:
                    g = self.gcd(f, h)
                    h //= g

                    if not h.is_ground:
                        factors.append((h, i*n))

                    f //= g
                    h = g
                    i += 1

            n *= p

            g = self.zero
            for monom, coeff in f.items():
                g[(_ // p for _ in monom)] = coeff**m
            f = g

        return factors

    def _rr_yun0_sqf_list(self, f):
        """Compute square-free decomposition of ``f`` in zero-characteristic ring ``K[X]``.

        References
        ==========

        * :cite:`Geddes1992algorithms`, algorithm 8.2
        * :cite:`LeeM2013factor`, algorithm 3.1

        """
        if f.is_ground:
            return []

        result, count = [], 1
        qs = [f.diff(x) for x in self.gens]

        g = f
        for q in qs:
            g = self.gcd(g, q)

        while f != 1:
            qs = [q // g for q in qs]
            f //= g
            qs = [q - f.diff(x) for x, q in zip(self.gens, qs)]

            g = f
            for q in qs:
                g = self.gcd(g, q)
            if g != 1:
                result.append((g, count))

            count += 1

        return result

    def is_squarefree(self, f):
        """
        Return ``True`` if ``f`` is a square-free polynomial in ``K[X]``.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)

        >>> ((x + y)**2).is_squarefree
        False
        >>> (x**2 + y**2).is_squarefree
        True

        """
        if f.is_ground:
            return True
        else:
            g = f
            for x in self.gens:
                g = self.gcd(g, f.diff(x))
                if g.is_ground:
                    return True
            return False

    def sqf_part(self, f):
        """
        Returns square-free part of a polynomial in ``K[X]``.

        Examples
        ========

        >>> R, x, y = ring('x y', ZZ)

        >>> R.sqf_part(x**3 + 2*x**2*y + x*y**2)
        x**2 + x*y

        """
        domain = self.domain

        if domain.is_FiniteField:
            g = self.one
            for f, _ in self.sqf_list(f)[1]:
                g *= f

            return g

        if not f:
            return f

        gcd = f
        for x in self.gens:
            gcd = self.gcd(gcd, f.diff(x))
        sqf = f // gcd

        if domain.is_Field:
            return sqf.monic()
        else:
            return sqf.primitive()[1]

    def sqf_norm(self, f):
        """
        Square-free norm of ``f`` in ``K[X]``, useful over algebraic domains.

        Returns ``s``, ``f``, ``r``, such that ``g(x) = f(x-sa)`` and ``r(x) = Norm(g(x))``
        is a square-free polynomial over K, where ``a`` is the algebraic extension of ``K``.

        Examples
        ========

        >>> _, x, y = ring('x y', QQ.algebraic_field(I))

        >>> (x*y + y**2).sqf_norm()
        (1, x*y - I*x + y**2 - 3*I*y - 2,
         x**2*y**2 + x**2 + 2*x*y**3 + 2*x*y + y**4 + 5*y**2 + 4)

        """
        domain = self.domain

        if not domain.is_AlgebraicField:
            raise DomainError(f'ground domain must be algebraic, got {domain}')

        new_ring = self.to_ground().inject(*domain.symbols, front=True)
        g = domain.mod.set_ring(new_ring)
        s = 0

        while True:
            h = f.inject(front=True)
            r = g.resultant(h)

            if r.is_squarefree:
                return s, f, r
            else:
                f = f.compose({x: x - domain.unit for x in self.gens})
                s += 1



"""Euclidean algorithms, GCDs, LCMs and polynomial remainder sequences."""

import math
import operator

from ..config import query
from ..ntheory import nextprime
from ..ntheory.modular import crt, symmetric_residue
from .polyerrors import HeuristicGCDFailed, HomomorphismFailed


class _GCD:
    """Mixin class for computing gcd."""

    def gcd(self, f, g):
        """Returns GCD of ``f`` and ``g``."""
        if not f and not g:
            return self.zero
        elif not f:
            return self._gcd_zero(g)
        elif not g:
            return self._gcd_zero(f)
        elif f.is_term:
            return self._gcd_term(f, g)
        elif g.is_term:
            return self._gcd_term(g, f)

        J, (f, g) = self._deflate(f, g)
        h = self._gcd(f, g)

        return self._inflate(h, J)

    def lcm(self, f, g):
        """Returns LCM of ``f`` and ``g``."""
        domain = self.domain

        if not domain.is_Field:
            fc, f = f.primitive()
            gc, g = g.primitive()
            c = domain.lcm(fc, gc)

        h = (f*g)//f.gcd(g)

        if not domain.is_Field:
            return h*c
        else:
            return h.monic()

    def _deflate(self, *polys):
        J = [0]*self.ngens

        for p in polys:
            for monom in p:
                for i, m in enumerate(monom):
                    J[i] = math.gcd(J[i], m)

        for i, b in enumerate(J):
            if not b:
                J[i] = 1

        J = tuple(J)

        if all(b == 1 for b in J):
            return J, polys

        H = []

        for p in polys:
            h = self.zero

            for I, coeff in p.items():
                N = [i//j for i, j in zip(I, J)]
                h[N] = coeff

            H.append(h)

        return J, H

    def _inflate(self, f, J):
        poly = self.zero

        for I, coeff in f.items():
            N = [i*j for i, j in zip(I, J)]
            poly[N] = coeff

        return poly

    def _gcd_zero(self, f):
        if self.domain.is_Field:
            return f.monic()
        else:
            return f if self.is_normal(f) else -f

    def _gcd_term(self, f, g):
        domain = self.domain
        ground_gcd = domain.gcd
        _mgcd, _cgcd = f.LT
        if domain.is_Field:
            for mg, cg in g.items():
                _mgcd = _mgcd.gcd(mg)
            _cgcd = domain.one
        else:
            for mg, cg in g.items():
                _mgcd = _mgcd.gcd(mg)
                _cgcd = ground_gcd(_cgcd, cg)
        return self.term_new(_mgcd, _cgcd)

    def _gcd(self, f, g):
        domain = self.domain

        if domain.is_RationalField:
            return self._gcd_QQ(f, g)
        elif domain.is_IntegerRing:
            return self._gcd_ZZ(f, g)
        elif domain.is_AlgebraicField:
            return self._gcd_AA(f, g)
        elif not domain.is_Exact:
            exact = domain.get_exact()
            ring = self.clone(domain=exact)
            f, g = map(operator.methodcaller('set_domain', exact), (f, g))
            h = ring._gcd(f, g)
            return h.set_domain(domain)
        elif domain.is_Field:
            return self._ff_prs_gcd(f, g)
        else:
            return self._rr_prs_gcd(f, g)

    def _gcd_ZZ(self, f, g):
        from .modulargcd import modgcd

        if query('USE_HEU_GCD'):
            try:
                return self._zz_heu_gcd(f, g)
            except HeuristicGCDFailed:
                pass

        _gcd_zz_methods = {'modgcd': modgcd,
                           'prs': self._rr_prs_gcd}

        method = _gcd_zz_methods[query('FALLBACK_GCD_ZZ_METHOD')]
        return method(f, g)

    def _gcd_QQ(self, f, g):
        domain = self.domain

        _, f = f.clear_denoms(convert=True)
        _, g = g.clear_denoms(convert=True)

        ring = self.clone(domain=domain.ring)

        h = ring._gcd_ZZ(f, g)
        h = h.set_ring(self)

        return h.monic()

    def _gcd_AA(self, f, g):
        from .modulargcd import func_field_modgcd

        _gcd_aa_methods = {'modgcd': func_field_modgcd,
                           'prs': self._ff_prs_gcd}

        method = _gcd_aa_methods[query('GCD_AA_METHOD')]
        return method(f, g)

    def _zz_heu_gcd(self, f, g):
        """
        Heuristic polynomial GCD in ``Z[X]``.

        Given univariate polynomials ``f`` and ``g`` in ``Z[X]``, returns
        their GCD, i.e. polynomial ``h``::

              h = gcd(f, g)

        The algorithm is purely heuristic which means it may fail to compute
        the GCD. This will be signaled by raising an exception. In this case
        you will need to switch to another GCD method.

        The algorithm computes the polynomial GCD by evaluating polynomials
        ``f`` and ``g`` at certain points and computing (fast) integer GCD
        of those evaluations. The polynomial GCD is recovered from the integer
        image by interpolation. The evaluation proces reduces f and g variable
        by variable into a large integer. The final step is to verify if the
        interpolated polynomial is the correct GCD.

        References
        ==========

        * :cite:`Liao1995heuristic`

        """
        assert self == f.ring == g.ring
        assert self.domain.is_IntegerRing

        x0 = self.gens[0]
        domain = self.domain

        fc = f.content()
        gc = g.content()

        gcd = self.domain.gcd(fc, gc)

        f = f.quo_ground(gcd)
        g = g.quo_ground(gcd)

        f_norm = f.max_norm()
        g_norm = g.max_norm()

        B = domain(2*min(f_norm, g_norm) + 29)

        x = max(min(B, 99*domain.sqrt(B)),
                2*min(f_norm // abs(f.LC),
                      g_norm // abs(g.LC)) + 4)

        cofactors = domain.cofactors if self.is_univariate else self.drop(0).cofactors

        for _ in range(query('HEU_GCD_MAX')):
            ff = f.eval(x0, x)
            gg = g.eval(x0, x)

            if ff and gg:
                h, cff, cfg = cofactors(ff, gg)
                h = self._gcd_interpolate(h, x)
                h = h.primitive()[1]

                if not f % h and not g % h:
                    return h*gcd

                cff = self._gcd_interpolate(cff, x)

                h, r = divmod(f, cff)

                if not r and not g % h:
                    return h*gcd

                cfg = self._gcd_interpolate(cfg, x)

                h, r = divmod(g, cfg)

                if not r and not f % h:
                    return h*gcd

            x = 73794*x * domain.sqrt(domain.sqrt(x)) // 27011

        raise HeuristicGCDFailed('no luck')

    def _gcd_interpolate(self, h, x):
        """Interpolate polynomial GCD from integer GCD."""
        f, i = self.zero, 0
        X = self.gens[0]

        while h:
            g = h % x

            if self.is_univariate:
                g = symmetric_residue(g, x)

            h = (h - g) // x

            f += X**i*g
            i += 1

        if not self.is_normal(f):
            f = -f

        return f

    def _rr_prs_gcd(self, f, g):
        """Computes polynomial GCD using subresultants over a ring."""
        if self.is_multivariate:
            ring, f, g = map(operator.methodcaller('eject', *self.gens[1:]),
                             (self, f, g))
            h = ring._rr_prs_gcd(f, g)
            return h.inject()

        domain = self.domain

        fc, ff = f.primitive()
        gc, fg = g.primitive()

        h = ff.subresultants(fg)[-1]
        _, h = h.primitive()
        c = domain.gcd(fc, gc)

        return h*c

    def _ff_prs_gcd(self, f, g):
        """Computes polynomial GCD using subresultants over a field."""
        if self.is_multivariate:
            ring, F, G = map(operator.methodcaller('eject', *self.gens[1:]),
                             (self, f, g))

            fc, F = F.primitive()
            gc, G = G.primitive()

            F, G = map(operator.methodcaller('inject'), (F, G))

            h = F.subresultants(G)[-1]
            c = ring.domain._ff_prs_gcd(fc, gc)
            h = h.eject(*self.gens[1:])
            _, h = h.primitive()
            h = h.inject()
            h *= c

            return h.monic()

        h = f.subresultants(g)[-1]

        return h.monic()

    def _primitive_prs(self, f, g):
        """
        Subresultant PRS algorithm in `K[X]`.

        Computes the last non-zero scalar subresultant of `f` and `g`
        and subresultant polynomial remainder sequence (PRS).

        The first subdeterminant is set to 1 by convention to match
        the polynomial and the scalar subdeterminants.
        If 'deg(f) < deg(g)', the subresultants of '(g,f)' are computed.

        Examples
        ========

        >>> _, x = ring('x', ZZ)

        >>> (x**2 + 1).resultant(x**2 - 1, includePRS=True)
        (4, [x**2 + 1, x**2 - 1, -2])

        References
        ==========

        * :cite:`Brown1978prs`
        * :cite:`Geddes1992algorithms`, example 7.6

        """
        domain = self.domain

        if self.is_multivariate:
            ring, f, g = map(operator.methodcaller('eject', *self.gens[1:]),
                             (self, f, g))
            res = ring._primitive_prs(f, g)
            return res[0], [_.inject() for _ in res[1]]

        n = f.degree()
        m = g.degree()

        if n < m:
            res, sub = self._primitive_prs(g, f)
            if res:
                res *= (-domain.one)**(n*m)
            return res, sub

        c, r = domain.zero, []

        if not f:
            return c, r

        r.append(f)

        if not g:
            return c, r

        r.append(g)
        d = n - m

        h = f.prem(g)
        h *= (-domain.one)**(d + 1)

        lc = g.LC
        c = -lc**d

        while h:
            k = h.degree()
            r.append(h)

            f, g, m, d = g, h, k, m - k

            h = f.prem(g)
            h = h.quo_ground(-lc*c**d)

            lc = g.LC
            c = domain.quo((-lc)**d, c**(d - 1))

        if r[-1].degree() > 0:
            c = domain.zero
        else:
            c = -c

        return c, r

    def _collins_resultant(self, f, g):
        """
        Collins's modular resultant algorithm in `ZZ[X]` or `QQ[X]`.

        References
        ==========

        * :cite:`Collins1971mod`, algorithm PRES

        """
        domain = self.domain

        if not f or not g:
            return self.drop(0).zero

        n = f.degree()
        m = g.degree()

        if domain.is_RationalField:
            cf, f = f.clear_denoms(convert=True)
            cg, g = g.clear_denoms(convert=True)

            ring = self.clone(domain=domain.ring)
            r = ring._collins_resultant(f, g)
            r = r.set_domain(domain)

            c = cf**n * cg**m
            c = domain.convert(c, domain.ring)

            return r.quo_ground(c)

        assert domain.is_IntegerRing

        A = f.max_norm()
        B = g.max_norm()

        a, b = f.LC, g.LC

        B = domain(2)*domain.factorial(domain(n + m))*A**m*B**n
        new_ring = self.drop(0)
        r, p, P = new_ring.zero, domain.one, domain.one

        while P <= B:
            while True:
                p = domain(nextprime(p))
                if a % p and b % p:
                    break

            p_domain = domain.finite_field(p)
            F, G = map(operator.methodcaller('set_domain', p_domain), (f, g))

            try:
                R = self.clone(domain=p_domain)._modular_resultant(F, G)
            except HomomorphismFailed:
                continue

            if P == 1:
                r = R
            else:
                def _crt(r, R):
                    return domain(crt([P, p], map(domain.convert, [r, R]),
                                  check=False, symmetric=True)[0])

                if new_ring.is_PolynomialRing:
                    r_new = new_ring.zero

                    for monom in r | R:
                        r_new[monom] = _crt(r.get(monom, 0), R.get(monom, 0))
                    r = r_new
                else:
                    r = _crt(r, R)

            P *= p

        return r

    def _modular_resultant(self, f, g):
        """
        Compute resultant of `f` and `g` in `GF(p)[X]`.

        References
        ==========

        * :cite:`Collins1971mod`, algorithm CPRES

        """
        domain = self.domain

        assert domain.is_FiniteField

        if self.is_univariate:
            return self._primitive_prs(f, g)[0]

        n = f.degree()
        m = g.degree()

        N = f.degree(1)
        M = g.degree(1)

        B = n*M + m*N

        new_ring = self.drop(0)
        r = new_ring.zero
        D = self.eject(1).domain.one
        domain_elts = iter(range(domain.order))

        while D.degree() <= B:
            while True:
                try:
                    a = next(domain_elts)
                except StopIteration as exc:
                    raise HomomorphismFailed('no luck') from exc

                F = f.eval(x=1, a=a)

                if F.degree() == n:
                    G = g.eval(x=1, a=a)

                    if G.degree() == m:
                        break

            R = self.drop(1)._modular_resultant(F, G)
            e = r.eval(x=0, a=a)

            if new_ring.is_univariate:
                R = new_ring.ground_new(R)
                e = new_ring.ground_new(e)
            else:
                R = R.set_ring(new_ring)
                e = e.set_ring(new_ring)

            d = D * D.eval(x=0, a=a)**-1
            d = d.set_ring(new_ring)
            r += d*(R - e)
            D *= D.ring.gens[0] - a

        return r



"""Inference in propositional logic"""

from ..core.sympify import sympify
from ..utilities import ordered
from .boolalg import And, Not, to_cnf


def satisfiable(expr, algorithm='dpll2', all_models=False):
    """
    Check satisfiability of a propositional sentence.
    Returns a model when it succeeds.
    Returns {true: true} for trivially true expressions.

    On setting all_models to True, if given expr is satisfiable then
    returns a generator of models. However, if expr is unsatisfiable
    then returns a generator containing the single element False.

    Examples
    ========

    >>> satisfiable(a & ~b)
    {a: True, b: False}
    >>> satisfiable(a & ~a)
    False
    >>> satisfiable(True)
    {true: True}
    >>> next(satisfiable(a & ~a, all_models=True))
    False
    >>> models = satisfiable((a >> b) & b, all_models=True)
    >>> next(models)
    {a: False, b: True}
    >>> next(models)
    {a: True, b: True}
    >>> def use_models(models):
    ...     for model in models:
    ...         if model:
    ...             # Do something with the model.
    ...             return model
    ...         else:
    ...             # Given expr is unsatisfiable.
    ...             print('UNSAT')
    >>> use_models(satisfiable(a >> ~a, all_models=True))
    {a: False}
    >>> use_models(satisfiable(a ^ a, all_models=True))
    UNSAT

    """
    expr = to_cnf(expr)
    if algorithm == 'dpll':
        from .algorithms.dpll import dpll_satisfiable
        return dpll_satisfiable(expr)
    elif algorithm == 'dpll2':
        from .algorithms.dpll2 import dpll_satisfiable
        return dpll_satisfiable(expr, all_models)
    else:
        raise NotImplementedError


def valid(expr):
    """
    Check validity of a propositional sentence.
    A valid propositional sentence is True under every assignment.

    Examples
    ========

    >>> valid(a | ~a)
    True
    >>> valid(a | b)
    False

    References
    ==========

    * https://en.wikipedia.org/wiki/Validity

    """
    return not satisfiable(Not(expr))


def pl_true(expr, model={}, deep=False):
    """
    Returns whether the given assignment is a model or not.

    If the assignment does not specify the value for every proposition,
    this may return None to indicate 'not obvious'.

    Parameters
    ==========

    model : dict, optional, default: {}
        Mapping of symbols to boolean values to indicate assignment.
    deep: boolean, optional, default: False
        Gives the value of the expression under partial assignments
        correctly. May still return None to indicate 'not obvious'.


    Examples
    ========

    >>> pl_true(a & b, {a: True, b: True})
    True
    >>> pl_true(a & b, {a: False})
    False
    >>> pl_true(a & b, {a: True})
    >>> pl_true(a & b, {a: True}, deep=True)
    >>> pl_true(a >> (b >> a))
    >>> pl_true(a >> (b >> a), deep=True)
    True
    >>> pl_true(a & ~a)
    >>> pl_true(a & ~a, deep=True)
    False
    >>> pl_true(a & b & (~a | ~b), {a: True})
    >>> pl_true(a & b & (~a | ~b), {a: True}, deep=True)
    False

    """
    from ..core import Symbol
    from .boolalg import BooleanFunction
    boolean = (True, False)

    def _validate(expr):
        if isinstance(expr, Symbol) or expr in boolean:
            return True
        if not isinstance(expr, BooleanFunction):
            return False
        return all(_validate(arg) for arg in expr.args)

    if expr in boolean:
        return expr
    expr = sympify(expr)
    if not _validate(expr):
        raise ValueError(f'{expr} is not a valid boolean expression')
    model = {k: v for k, v in model.items() if v in boolean}
    result = expr.subs(model)
    if result in boolean:
        return bool(result)
    if deep:
        model = {k: True for k in result.atoms()}
        if pl_true(result, model):
            if valid(result):
                return True
        else:
            if not satisfiable(result):
                return False


def entails(expr, formula_set={}):
    """
    Check whether the given expr_set entail an expr.
    If formula_set is empty then it returns the validity of expr.

    Examples
    ========

    >>> entails(a, [a >> b, b >> c])
    False
    >>> entails(c, [a >> b, b >> c, a])
    True
    >>> entails(a >> b)
    False
    >>> entails(a >> (b >> a))
    True

    References
    ==========

    * https://en.wikipedia.org/wiki/Logical_consequence

    """
    formula_set = list(formula_set)
    formula_set.append(Not(expr))
    return not satisfiable(And(*formula_set))


class KB:
    """Base class for all knowledge bases."""

    def __init__(self, sentence=None):
        self.clauses_ = set()
        if sentence:
            self.tell(sentence)

    def tell(self, sentence):
        raise NotImplementedError

    def ask(self, query):
        raise NotImplementedError

    def retract(self, sentence):
        raise NotImplementedError

    @property
    def clauses(self):
        return list(ordered(self.clauses_))


class PropKB(KB):
    """A KB for Propositional Logic.  Inefficient, with no indexing."""

    def tell(self, sentence):
        """Add the sentence's clauses to the KB

        Examples
        ========

        >>> l = PropKB()
        >>> l.clauses
        []

        >>> l.tell(x | y)
        >>> l.clauses
        [x | y]

        >>> l.tell(y)
        >>> l.clauses
        [y, x | y]

        """
        for c in And.make_args(to_cnf(sentence)):
            self.clauses_.add(c)

    def ask(self, query):
        """Checks if the query is true given the set of clauses.

        Examples
        ========

        >>> l = PropKB()
        >>> l.tell(x & ~y)
        >>> l.ask(x)
        True
        >>> l.ask(y)
        False

        """
        return entails(query, self.clauses_)

    def retract(self, sentence):
        """Remove the sentence's clauses from the KB

        Examples
        ========

        >>> l = PropKB()
        >>> l.clauses
        []

        >>> l.tell(x | y)
        >>> l.clauses
        [x | y]

        >>> l.retract(x | y)
        >>> l.clauses
        []

        """
        for c in And.make_args(to_cnf(sentence)):
            self.clauses_.discard(c)



"""Algorithms for partial fraction decomposition of rational functions."""

import itertools

from ..core import Add, Dummy, Function, Integer, Lambda, preorder_traversal
from ..core.sympify import sympify
from ..utilities import numbered_symbols
from . import Poly, RootSum, cancel, factor
from .polyerrors import PolynomialError
from .polyoptions import allowed_flags, set_defaults
from .polytools import parallel_poly_from_expr


__all__ = 'apart', 'apart_list', 'assemble_partfrac_list'


def apart(f, x=None, full=False, **options):
    """
    Compute partial fraction decomposition of a rational function.

    Given a rational function ``f``, computes the partial fraction
    decomposition of ``f``. Two algorithms are available: One is based on the
    undertermined coefficients method, the other is Bronstein's full partial
    fraction decomposition algorithm.

    The undetermined coefficients method (selected by ``full=False``) uses
    polynomial factorization (and therefore accepts the same options as
    factor) for the denominator. Per default it works over the rational
    numbers, therefore decomposition of denominators with non-rational roots
    (e.g. irrational, complex roots) is not supported by default (see options
    of factor).

    Bronstein's algorithm can be selected by using ``full=True`` and allows a
    decomposition of denominators with non-rational roots. A human-readable
    result can be obtained via ``doit()`` (see examples below).

    Examples
    ========

    By default, using the undetermined coefficients method:

    >>> apart(y/(x + 2)/(x + 1), x)
    -y/(x + 2) + y/(x + 1)

    The undetermined coefficients method does not provide a result when the
    denominators roots are not rational:

    >>> apart(y/(x**2 + x + 1), x)
    y/(x**2 + x + 1)

    You can choose Bronstein's algorithm by setting ``full=True``:

    >>> apart(y/(x**2 + x + 1), x, full=True)
    RootSum(_w**2 + _w + 1, Lambda(_a, (-2*y*_a/3 - y/3)/(x - _a)))

    Calling ``doit()`` yields a human-readable result:

    >>> apart(y/(x**2 + x + 1), x, full=True).doit()
    (-y/3 - 2*y*(-1/2 - sqrt(3)*I/2)/3)/(x + 1/2 + sqrt(3)*I/2) + (-y/3 -
        2*y*(-1/2 + sqrt(3)*I/2)/3)/(x + 1/2 - sqrt(3)*I/2)

    See Also
    ========

    apart_list, assemble_partfrac_list

    """
    allowed_flags(options, [])

    f = sympify(f)

    if f.is_Atom:
        return f
    else:
        P, Q = f.as_numer_denom()

    _options = options.copy()
    options = set_defaults(options, extension=True)
    try:
        (P, Q), _ = parallel_poly_from_expr((P, Q), x, **options)
    except PolynomialError as exc:
        if f.is_commutative:
            raise PolynomialError(exc) from exc
        # non-commutative
        if f.is_Mul:
            c, nc = f.args_cnc(split_1=False)
            nc = f.func(*[apart(i, x=x, full=full, **_options) for i in nc])
            if c:
                c = apart(f.func._from_args(c), x=x, full=full, **_options)
                return c*nc
            else:
                return nc
        elif f.is_Add:
            c = []
            nc = []
            for i in f.args:
                if i.is_commutative:
                    c.append(i)
                else:
                    nc.append(apart(i, x=x, full=full, **_options))
            return apart(f.func(*c), x=x, full=full, **_options) + f.func(*nc)
        else:
            reps = []
            pot = preorder_traversal(f)
            next(pot)
            for e in pot:
                reps.append((e, apart(e, x=x, full=full, **_options)))
                pot.skip()  # this was handled successfully
            return f.xreplace(dict(reps))

    if P.is_multivariate:
        fc = f.cancel()
        if fc != f:
            return apart(fc, x=x, full=full, **_options)

        raise NotImplementedError(
            'multivariate partial fraction decomposition')

    common, P, Q = P.cancel(Q)

    poly, P = P.div(Q, auto=True)
    P, Q = P.rat_clear_denoms(Q)

    if Q.degree() <= 1:
        partial = P/Q
    else:
        if not full:
            partial = apart_undetermined_coeffs(P, Q)
        else:
            partial = apart_full_decomposition(P, Q)

    terms = Integer(0)

    for term in Add.make_args(partial):
        if term.has(RootSum):
            terms += term
        else:
            terms += factor(term)

    return common*(poly.as_expr() + terms)


def apart_undetermined_coeffs(P, Q):
    """Partial fractions via method of undetermined coefficients."""
    X = numbered_symbols(cls=Dummy)
    partial, symbols = [], []

    _, factors = Q.factor_list()

    for f, k in factors:
        n, q = f.degree(), Q

        for i in range(1, k + 1):
            coeffs, q = list(itertools.islice(X, n)), q.quo(f)
            partial.append((coeffs, q, f, i))
            symbols.extend(coeffs)

    dom = Q.domain.inject(*symbols)
    F = Integer(0).as_poly(Q.gen, domain=dom)

    for i, (coeffs, q, f, k) in enumerate(partial):
        h = Poly(coeffs, Q.gen, domain=dom)
        partial[i] = (h, f, k)
        q = q.set_domain(dom)
        F += h*q

    system, result = [], Integer(0)

    for (k,), coeff in F.terms():
        system.append(coeff - P.coeff_monomial((k,)))

    from ..solvers import solve
    solution = solve(system, symbols)[0]

    for h, f, k in partial:
        h = h.as_expr().subs(solution)
        result += h/f.as_expr()**k

    return result


def apart_full_decomposition(P, Q):
    """
    Bronstein's full partial fraction decomposition algorithm.

    Given a univariate rational function ``f``, performing only GCD
    operations over the algebraic closure of the initial ground domain
    of definition, compute full partial fraction decomposition with
    fractions having linear denominators.

    Note that no factorization of the initial denominator of ``f`` is
    performed. The final decomposition is formed in terms of a sum of
    :class:`RootSum` instances.

    References
    ==========

    * :cite:`Bronstein1993partial`

    """
    return assemble_partfrac_list(apart_list(P/Q, P.gens[0]))


def apart_list(f, x=None, dummies=None, **options):
    """
    Compute partial fraction decomposition of a rational function
    and return the result in structured form.

    Given a rational function ``f`` compute the partial fraction decomposition
    of ``f``. Only Bronstein's full partial fraction decomposition algorithm
    is supported by this method. The return value is highly structured and
    perfectly suited for further algorithmic treatment rather than being
    human-readable. The function returns a tuple holding three elements:

    * The first item is the common coefficient, free of the variable `x` used
      for decomposition. (It is an element of the base field `K`.)

    * The second item is the polynomial part of the decomposition. This can be
      the zero polynomial. (It is an element of `K[x]`.)

    * The third part itself is a list of quadruples. Each quadruple
      has the following elements in this order:

      - The (not necessarily irreducible) polynomial `D` whose roots `w_i` appear
        in the linear denominator of a bunch of related fraction terms. (This item
        can also be a list of explicit roots. However, at the moment ``apart_list``
        never returns a result this way, but the related ``assemble_partfrac_list``
        function accepts this format as input.)

      - The numerator of the fraction, written as a function of the root `w`

      - The linear denominator of the fraction *excluding its power exponent*,
        written as a function of the root `w`.

      - The power to which the denominator has to be raised.

    On can always rebuild a plain expression by using the function ``assemble_partfrac_list``.

    Examples
    ========

    A first example:

    >>> f = (2*x**3 - 2*x) / (x**2 - 2*x + 1)
    >>> pfd = apart_list(f)
    >>> pfd
    (1,
    Poly(2*x + 4, x, domain='ZZ'),
    [(Poly(_w - 1, _w, domain='ZZ'), Lambda(_a, 4), Lambda(_a, x - _a), 1)])

    >>> assemble_partfrac_list(pfd)
    2*x + 4 + 4/(x - 1)

    Second example:

    >>> f = (-2*x - 2*x**2) / (3*x**2 - 6*x)
    >>> pfd = apart_list(f)
    >>> pfd
    (-1,
    Poly(2/3, x, domain='QQ'),
    [(Poly(_w - 2, _w, domain='ZZ'), Lambda(_a, 2), Lambda(_a, x - _a), 1)])

    >>> assemble_partfrac_list(pfd)
    -2/3 - 2/(x - 2)

    Another example, showing symbolic parameters:

    >>> pfd = apart_list(t/(x**2 + x + t), x)
    >>> pfd
    (1,
    Poly(0, x, domain='ZZ[t]'),
    [(Poly(_w**2 + _w + t, _w, domain='ZZ[t]'),
    Lambda(_a, -2*t*_a/(4*t - 1) - t/(4*t - 1)),
    Lambda(_a, x - _a), 1)])

    >>> assemble_partfrac_list(pfd)
    RootSum(t + _w**2 + _w, Lambda(_a, (-2*t*_a/(4*t - 1) - t/(4*t - 1))/(x - _a)))

    This example is taken from Bronstein's original paper:

    >>> f = 36 / (x**5 - 2*x**4 - 2*x**3 + 4*x**2 + x - 2)
    >>> pfd = apart_list(f)
    >>> pfd
    (1,
    Poly(0, x, domain='ZZ'),
    [(Poly(_w - 2, _w, domain='ZZ'), Lambda(_a, 4), Lambda(_a, x - _a), 1),
    (Poly(_w**2 - 1, _w, domain='ZZ'), Lambda(_a, -3*_a - 6), Lambda(_a, x - _a), 2),
    (Poly(_w + 1, _w, domain='ZZ'), Lambda(_a, -4), Lambda(_a, x - _a), 1)])

    >>> assemble_partfrac_list(pfd)
    -4/(x + 1) - 3/(x + 1)**2 - 9/(x - 1)**2 + 4/(x - 2)

    See also
    ========

    apart, assemble_partfrac_list

    References
    ==========

    * :cite:`Bronstein1993partial`

    """
    allowed_flags(options, [])

    f = sympify(f)

    if f.is_Atom:
        return f
    else:
        P, Q = f.as_numer_denom()

    options = set_defaults(options, extension=True)
    (P, Q), _ = parallel_poly_from_expr((P, Q), x, **options)

    if P.is_multivariate:
        raise NotImplementedError('multivariate partial fraction decomposition')

    common, P, Q = P.cancel(Q)

    poly, P = P.div(Q, auto=True)
    P, Q = P.rat_clear_denoms(Q)

    polypart = poly

    if dummies is None:
        def dummies(name):
            d = Dummy(name)
            while True:
                yield d

        dummies = dummies('w')

    rationalpart = apart_list_full_decomposition(P, Q, dummies)

    return common, polypart, rationalpart


def apart_list_full_decomposition(P, Q, dummygen):
    """
    Bronstein's full partial fraction decomposition algorithm.

    Given a univariate rational function ``f``, performing only GCD
    operations over the algebraic closure of the initial ground domain
    of definition, compute full partial fraction decomposition with
    fractions having linear denominators.

    Note that no factorization of the initial denominator of ``f`` is
    performed. The final decomposition is formed in terms of a sum of
    :class:`RootSum` instances.

    References
    ==========

    * :cite:`Bronstein1993partial`

    """
    f, x, U = P/Q, P.gen, []

    u = Function('u')(x)
    a = Dummy('a')

    Q_c, Q_sqf = Q.sqf_list()
    if Q_c != 1 and Q_sqf:
        if Q_sqf[0][1] == 1:
            Q_sqf[0] = Q_c*Q_sqf[0][0], 1
        else:
            Q_sqf.insert(0, (Q_c.as_poly(x), 1))

    partial = []

    for d, n in Q_sqf:
        b = d.as_expr()
        U += [u.diff((x, n - 1))]

        h = cancel(f*b**n) / u**n

        H, subs = [h], []

        for j in range(1, n):
            H += [H[-1].diff(x) / j]

        for j in range(1, n + 1):
            subs += [(U[j - 1], b.diff((x, j)) / j)]

        for j in range(n):
            P, Q = cancel(H[j]).as_numer_denom()

            for i in range(j + 1):
                P = P.subs([subs[j - i]])

            Q = Q.subs([subs[0]])

            P = P.as_poly(x)
            Q = Q.as_poly(x)

            G = P.gcd(d)
            D = d.quo(G)

            B, g = Q.half_gcdex(D)
            b = (P * B.quo(g)).rem(D)

            Dw = D.subs({x: next(dummygen)})
            numer = Lambda(a, b.as_expr().subs({x: a}))
            denom = Lambda(a, (x - a))
            exponent = n-j

            partial.append((Dw, numer, denom, exponent))

    return partial


def assemble_partfrac_list(partial_list):
    r"""Reassemble a full partial fraction decomposition
    from a structured result obtained by the function ``apart_list``.

    Examples
    ========

    This example is taken from Bronstein's original paper:

    >>> f = 36 / (x**5 - 2*x**4 - 2*x**3 + 4*x**2 + x - 2)
    >>> pfd = apart_list(f)
    >>> pfd
    (1,
    Poly(0, x, domain='ZZ'),
    [(Poly(_w - 2, _w, domain='ZZ'), Lambda(_a, 4), Lambda(_a, x - _a), 1),
    (Poly(_w**2 - 1, _w, domain='ZZ'), Lambda(_a, -3*_a - 6), Lambda(_a, x - _a), 2),
    (Poly(_w + 1, _w, domain='ZZ'), Lambda(_a, -4), Lambda(_a, x - _a), 1)])

    >>> assemble_partfrac_list(pfd)
    -4/(x + 1) - 3/(x + 1)**2 - 9/(x - 1)**2 + 4/(x - 2)

    If we happen to know some roots we can provide them easily inside the structure:

    >>> pfd = apart_list(2/(x**2-2))
    >>> pfd
    (1,
    Poly(0, x, domain='ZZ'),
    [(Poly(_w**2 - 2, _w, domain='ZZ'),
    Lambda(_a, _a/2), Lambda(_a, x - _a),
    1)])

    >>> pfda = assemble_partfrac_list(pfd)
    >>> pfda
    RootSum(_w**2 - 2, Lambda(_a, _a/(x - _a)))/2

    >>> pfda.doit()
    -sqrt(2)/(2*(x + sqrt(2))) + sqrt(2)/(2*(x - sqrt(2)))

    >>> a = Dummy('a')
    >>> pfd = (1, Integer(0).as_poly(x),
    ...        [([sqrt(2), -sqrt(2)],
    ...          Lambda(a, a/2), Lambda(a, -a + x), 1)])

    >>> assemble_partfrac_list(pfd)
    -sqrt(2)/(2*(x + sqrt(2))) + sqrt(2)/(2*(x - sqrt(2)))

    See also
    ========

    apart, apart_list

    """
    # Common factor
    common = partial_list[0]

    # Polynomial part
    polypart = partial_list[1]
    pfd = polypart.as_expr()

    # Rational parts
    for r, nf, df, ex in partial_list[2]:
        if isinstance(r, Poly):
            # Assemble in case the roots are given implicitly by a polynomials
            an, nu = nf.variables, nf.expr
            ad, de = df.variables, df.expr
            # Hack to make dummies equal because Lambda created new Dummies
            de = de.subs({ad[0]: an[0]})
            func = Lambda(an, nu/de**ex)
            pfd += RootSum(r, func, auto=False, quadratic=False)
        else:
            # Assemble in case the roots are given explicitly by a list of algebraic numbers
            for root in r:
                pfd += nf(root)/df(root)**ex

    return common*pfd



import pytest

from diofant import (Function, I, Mul, Rational, Symbol, cot, exp, factorial,
                     log, pi, residue, root, sin, sqrt, tan, tanh)
from diofant.abc import a, s, x, z


__all__ = ()


def test_basic1():
    assert residue(1/x, x, 0) == 1
    assert residue(-2/x, x, 0) == -2
    assert residue(81/x, x, 0) == 81
    assert residue(1/x**2, x, 0) == 0
    assert residue(0, x, 0) == 0
    assert residue(5, x, 0) == 0
    assert residue(x, x, 0) == 0
    assert residue(x**2, x, 0) == 0


def test_basic2():
    assert residue(1/x, x, 1) == 0
    assert residue(-2/x, x, 1) == 0
    assert residue(81/x, x, -1) == 0
    assert residue(1/x**2, x, 1) == 0
    assert residue(0, x, 1) == 0
    assert residue(5, x, 1) == 0
    assert residue(x, x, 1) == 0
    assert residue(x**2, x, 5) == 0


def test_f():
    f = Function('f')
    assert residue(f(x)/x**5, x, 0) == f(x).diff((x, 4)).subs({x: 0})/24


def test_functions():
    assert residue(1/sin(x), x, 0) == 1
    assert residue(2/sin(x), x, 0) == 2
    assert residue(1/sin(x)**2, x, 0) == 0
    assert residue(1/sin(x)**5, x, 0) == Rational(3, 8)


def test_expressions():
    assert residue(1/(x + 1), x, 0) == 0
    assert residue(1/(x + 1), x, -1) == 1
    assert residue(1/(x**2 + 1), x, -1) == 0
    assert residue(1/(x**2 + 1), x, I) == -I/2
    assert residue(1/(x**2 + 1), x, -I) == I/2
    assert residue(1/(x**4 + 1), x, 0) == 0
    assert residue(1/(x**4 + 1), x, exp(I*pi/4)) == -root(-1, 4)/4
    assert residue(1/(x**2 + a**2)**2, x, a*I) == -I/4/a**3


@pytest.mark.xfail
def test_expressions_failing():
    n = Symbol('n', integer=True, positive=True)
    assert residue(exp(z)/(z - pi*I/4*a)**n, z, I*pi*a) == \
        exp(I*pi*a/4)/factorial(n - 1)


def test_NotImplemented():
    pytest.raises(NotImplementedError, lambda: residue(exp(1/z), z, 0))


def test_bug():
    assert residue(2**(z)*(s + z)*(1 - s - z)/z**2, z, 0) == \
        1 + s*log(2) - s**2*log(2) - 2*s


def test_sympyissue_5654():
    assert residue(1/(x**2 + a**2)**2, x, a*I) == -I/(4*a**3)


def test_sympyissue_6499():
    assert residue(1/(exp(z) - 1), z, 0) == 1


def test_sympyissue_21177():
    e1 = cot(pi*x)/((x - 1)*(x - 2) + 1)
    e2 = cot(pi*x)/(x**2 - 3*x + 3)
    pt = Rational(3, 2) - sqrt(3)*I/2
    ans = -sqrt(3)*tanh(sqrt(3)*pi/2)/3

    assert residue(e1, x, pt) == ans
    assert residue(e2, x, pt) == ans


def test_sympyissue_21176():
    e = x**2*cot(pi*x)/(x**4 + 1)
    pt = -sqrt(2)/2 - sqrt(2)*I/2
    assert residue(e, x, pt) == sqrt(2)*I/Mul(2, -2 + 2*I, tan(sqrt(2)*pi/2 +
                                                               sqrt(2)*I*pi/2),
                                              evaluate=False)



"""
This module provides convenient functions to transform diofant expressions to
lambda functions which can be used to calculate numerical values very fast.
"""

from __future__ import annotations

import inspect
import textwrap
import typing

from ..core.compatibility import is_sequence, iterable
from ..external import import_module
from .decorator import doctest_depends_on


# XXX make pylint & flake8 happy
import_module  # pylint: disable=pointless-statement


# These are the namespaces the lambda functions will use.
MATH: dict[str, typing.Any] = {}
MPMATH: dict[str, typing.Any] = {}
NUMPY: dict[str, typing.Any] = {}
DIOFANT: dict[str, typing.Any] = {}

# Default namespaces, letting us define translations that can't be defined
# by simple variable maps, like I => 1j
# These are separate from the names above because the above names are modified
# throughout this file, whereas these should remain unmodified.
MATH_DEFAULT: dict[str, typing.Any] = {}
MPMATH_DEFAULT: dict[str, typing.Any] = {}
NUMPY_DEFAULT: dict[str, typing.Any] = {'I': 1j}
DIOFANT_DEFAULT: dict[str, typing.Any] = {}

# Mappings between diofant and other modules function names.
MATH_TRANSLATIONS = {
    'Abs': 'fabs',
    'ceiling': 'ceil',
    'E': 'e',
    'ln': 'log',
}

MPMATH_TRANSLATIONS = {
    'Abs': 'fabs',
    'elliptic_k': 'ellipk',
    'elliptic_f': 'ellipf',
    'elliptic_e': 'ellipe',
    'elliptic_pi': 'ellippi',
    'ceiling': 'ceil',
    'chebyshevt': 'chebyt',
    'chebyshevu': 'chebyu',
    'E': 'e',
    'I': 'j',
    'ln': 'log',
    # 'lowergamma': 'lower_gamma',
    'oo': 'inf',
    # 'uppergamma': 'upper_gamma',
    'LambertW': 'lambertw',
    'MutableDenseMatrix': 'matrix',
    'ImmutableMatrix': 'matrix',
    'conjugate': 'conj',
    'dirichlet_eta': 'altzeta',
    'Ei': 'ei',
    'Shi': 'shi',
    'Chi': 'chi',
    'Si': 'si',
    'Ci': 'ci',
    'Ynm': 'spherharm',
    'RisingFactorial': 'rf',
    'FallingFactorial': 'ff',
}

NUMPY_TRANSLATIONS = {
    'Abs': 'abs',
    'acos': 'arccos',
    'acosh': 'arccosh',
    'arg': 'angle',
    'asin': 'arcsin',
    'asinh': 'arcsinh',
    'atan': 'arctan',
    'atan2': 'arctan2',
    'atanh': 'arctanh',
    'ceiling': 'ceil',
    'E': 'e',
    'im': 'imag',
    'ln': 'log',
    'Mod': 'mod',
    'oo': 'inf',
    're': 'real',
    'SparseMatrix': 'array',
    'ImmutableSparseMatrix': 'array',
    'Matrix': 'array',
    'MutableDenseMatrix': 'array',
    'ImmutableMatrix': 'array',
    'ImmutableDenseMatrix': 'array',
}

# Available modules:
MODULES = {
    'math': (MATH, MATH_DEFAULT, MATH_TRANSLATIONS, ('from math import *',)),
    'mpmath': (MPMATH, MPMATH_DEFAULT, MPMATH_TRANSLATIONS, ('from mpmath import *',)),
    'numpy': (NUMPY, NUMPY_DEFAULT, NUMPY_TRANSLATIONS, ("import_module('numpy')",)),
    'diofant': (DIOFANT, DIOFANT_DEFAULT, {}, (
        'from diofant.functions import *',
        'from diofant.matrices import *',
        'from diofant import Sum, Integral, pi, oo, nan, zoo, E, I')),
}


def _import(module):
    """
    Creates a global translation dictionary for module.

    The argument module has to be one of the following strings: "math",
    "mpmath", "numpy", "diofant".
    These dictionaries map names of python functions to their equivalent in
    other modules.

    """
    try:
        namespace, namespace_default, translations, import_commands = MODULES[
            module]
    except KeyError as exc:
        raise NameError(f"'{module}' module can't be used "
                        'for lambdification') from exc

    # Clear namespace or exit
    if namespace != namespace_default:
        namespace.clear()
        namespace.update(namespace_default)

    for import_command in import_commands:
        if import_command.startswith('import_module'):
            module = eval(import_command)  # pylint: disable=eval-used

            if module is not None:
                namespace.update(module.__dict__)
                continue
        else:
            exec(import_command, {}, namespace)  # pylint: disable=exec-used
            continue

        raise ImportError(
            f"can't import '{module}' with '{import_command}' command")

    # Add translated names to namespace
    for diofantname, translation in translations.items():
        namespace[diofantname] = namespace[translation]


@doctest_depends_on(modules=('numpy'))
def lambdify(args, expr, modules=None, printer=None, use_imps=True,
             dummify=True):
    """
    Returns a lambda function for fast calculation of numerical values.

    If not specified differently by the user, ``modules`` defaults to
    ``["numpy"]`` if NumPy is installed, and ``["math", "mpmath", "sympy"]``
    if it isn't, that is, Diofant functions are replaced as far as possible by
    either ``numpy`` functions if available, and Python's standard library
    ``math``, or ``mpmath`` functions otherwise. To change this behavior, the
    "modules" argument can be used. It accepts:

     - the strings "math", "mpmath", "numpy", "diofant"
     - any modules (e.g. math)
     - dictionaries that map names of diofant functions to arbitrary functions
     - lists that contain a mix of the arguments above, with higher priority
       given to entries appearing first.

    The default behavior is to substitute all arguments in the provided
    expression with dummy symbols. This allows for applied functions (e.g.
    f(t)) to be supplied as arguments. Call the function with dummify=False if
    dummy substitution is unwanted (and `args` is not a string). If you want
    to view the lambdified function or provide "diofant" as the module, you
    should probably set dummify=False.

    In previous releases ``lambdify`` replaced ``Matrix`` with ``numpy.matrix``
    by default. As of release 0.7.7 ``numpy.array`` is the default.
    To get the old default behavior you must pass in ``[{'ImmutableMatrix':
    numpy.matrix}, 'numpy']`` to the ``modules`` kwarg.

    (1) Use one of the provided modules:

        >>> f = lambdify(x, sin(x), 'math')

        Attention: Functions that are not in the math module will throw a name
                   error when the lambda function is evaluated! So this would
                   be better:

        >>> f = lambdify(x, sin(x)*gamma(x), ('math', 'mpmath', 'diofant'))

    (2) Use some other module:

        >>> import numpy
        >>> f = lambdify((x, y), tan(x*y), numpy)

        Attention: There are naming differences between numpy and diofant. So if
                   you simply take the numpy module, e.g. diofant.atan will not be
                   translated to numpy.arctan. Use the modified module instead
                   by passing the string "numpy":

        >>> f = lambdify((x, y), tan(x*y), 'numpy')
        >>> f(1, 2)
        -2.18503986326
        >>> from numpy import array
        >>> f(array([1, 2, 3]), array([2, 3, 5]))
        [-2.18503986 -0.29100619 -0.8559934 ]

    (3) Use a dictionary defining custom functions:

        >>> def my_cool_function(x):
        ...     return f'sin({x}) is cool'
        >>> myfuncs = {'sin': my_cool_function}
        >>> f = lambdify(x, sin(x), myfuncs)
        >>> f(1)
        'sin(1) is cool'

    Examples
    ========

    >>> from diofant.abc import w

    >>> f = lambdify(x, x**2)
    >>> f(2)
    4
    >>> f = lambdify((x, y, z), [z, y, x])
    >>> f(1, 2, 3)
    [3, 2, 1]
    >>> f = lambdify(x, sqrt(x))
    >>> f(4)
    2.0
    >>> f = lambdify((x, y), sin(x*y)**2)
    >>> f(0, 5)
    0.0
    >>> row = lambdify((x, y), Matrix((x, x + y)).T, modules='diofant')
    >>> row(1, 2)
    Matrix([[1, 3]])

    Tuple arguments are handled and the lambdified function should
    be called with the same type of arguments as were used to create
    the function.:

    >>> f = lambdify((x, (y, z)), x + y)
    >>> f(1, (2, 4))
    3

    A more robust way of handling this is to always work with flattened
    arguments:

    >>> args = w, (x, (y, z))
    >>> vals = 1, (2, (3, 4))
    >>> f = lambdify(flatten(args), w + x + y + z)
    >>> f(*flatten(vals))
    10

    Functions present in `expr` can also carry their own numerical
    implementations, in a callable attached to the ``_imp_``
    attribute.  Usually you attach this using the
    ``implemented_function`` factory:

    >>> f = implemented_function(Function('f'), lambda x: x+1)
    >>> func = lambdify(x, f(x))
    >>> func(4)
    5

    ``lambdify`` always prefers ``_imp_`` implementations to implementations
    in other namespaces, unless the ``use_imps`` input parameter is False.

    """
    from ..core import Symbol
    from ..printing.lambdarepr import MpmathPrinter, NumPyPrinter
    from .iterables import flatten

    module_provided = True

    # If the user hasn't specified any modules, use what is available.
    if modules is None:
        module_provided = False

        try:
            _import('numpy')
        except ImportError:
            # Use either numpy (if available) or python.math where possible.
            # XXX: This leads to different behaviour on different systems and
            #      might be the reason for irreproducible errors.
            modules = ['math', 'mpmath', 'diofant']
        else:
            modules = ['numpy']

    # Get the needed namespaces.
    namespaces = []
    # First find any function implementations
    if use_imps:
        namespaces.append(_imp_namespace(expr))
    # Check for dict before iterating
    if isinstance(modules, (dict, str)) or not hasattr(modules, '__iter__'):
        namespaces.append(modules)
    else:
        namespaces += list(modules)
    # fill namespace with first having highest priority
    namespace = {}
    for m in namespaces[::-1]:
        buf = _get_namespace(m)
        namespace.update(buf)

    if hasattr(expr, 'atoms'):
        # Try if you can extract symbols from the expression.
        # Move on if expr.atoms in not implemented.
        syms = expr.atoms(Symbol)
        for term in syms:
            namespace.update({str(term): term})

    if 'numpy' in namespaces and printer is None:
        printer = NumPyPrinter

    if 'mpmath' in namespaces and printer is None:
        printer = MpmathPrinter

    # Get the names of the args, for creating a docstring
    if not iterable(args):
        args = args,
    names = []
    for n, var in enumerate(args):
        if hasattr(var, 'name'):
            names.append(var.name)
        else:
            # Cannot infer name with certainty. arg_# will have to do.
            names.append('arg_' + str(n))

    # Create lambda function.
    lstr = lambdastr(args, expr, printer=printer, dummify=dummify)
    flat = '__flatten_args__'

    if flat in lstr:
        namespace.update({flat: flatten})
    func = eval(lstr, namespace)  # pylint: disable=eval-used
    # For numpy lambdify, wrap all input arguments in arrays.
    if module_provided and 'numpy' in namespaces:
        def array_wrap(funcarg):
            def wrapper(*argsx, **kwargsx):
                return funcarg(*[namespace['asarray'](i) for i in argsx], **kwargsx)
            return wrapper
        func = array_wrap(func)
    # Apply the docstring
    sig = f"func({', '.join(str(i) for i in names)})"
    sig = textwrap.fill(sig, subsequent_indent=' '*8)
    expr_str = str(expr)
    if len(expr_str) > 78:
        expr_str = textwrap.wrap(expr_str, 75)[0] + '...'
    func.__doc__ = f"""Created with lambdify. Signature:

{sig}

Expression:

{expr_str}"""
    return func


def _get_namespace(m):
    """This is used by _lambdify to parse its arguments."""
    if isinstance(m, str):
        _import(m)
        return MODULES[m][0]
    elif isinstance(m, dict):
        return m
    elif hasattr(m, '__dict__'):
        return m.__dict__
    else:
        raise TypeError(f'Argument must be either a string, dict or module but it is: {m}')


def lambdastr(args, expr, printer=None, dummify=False):
    """
    Returns a string that can be evaluated to a lambda function.

    Examples
    ========

    >>> lambdastr(x, x**2)
    'lambda x: (x**2)'
    >>> lambdastr((x, y, z), [z, y, x])
    'lambda x,y,z: ([z, y, x])'

    Although tuples may not appear as arguments to lambda in Python 3,
    lambdastr will create a lambda function that will unpack the original
    arguments so that nested arguments can be handled:

    >>> lambdastr((x, (y, z)), x + y)
    'lambda _0,_1: (lambda x,y,z: (x + y))(*list(__flatten_args__([_0,_1])))'

    """
    # Transforming everything to strings.
    from ..core import Dummy, Function, Symbol, sympify
    from ..utilities import flatten

    if printer is not None:
        if inspect.isfunction(printer):
            lambdarepr = printer
        else:
            if inspect.isclass(printer):
                def lambdarepr(expr):
                    return printer().doprint(expr)
            else:
                def lambdarepr(expr):
                    return printer.doprint(expr)
    else:
        # XXX: This has to be done here because of circular imports
        from ..printing.lambdarepr import lambdarepr

    def sub_args(args, dummies_dict):
        if isinstance(args, str):
            return args
        elif iterable(args):
            dummies = flatten([sub_args(a, dummies_dict) for a in args])
            return ','.join(str(a) for a in dummies)
        else:
            # Sub in dummy variables for functions or symbols
            if isinstance(args, (Function, Symbol)):
                dummies = Dummy()
                dummies_dict.update({args: dummies})
                return lambdarepr(dummies)
            else:
                return lambdarepr(args)

    def sub_expr(expr, dummies_dict):
        try:
            expr = sympify(expr).xreplace(dummies_dict)
        except (TypeError, AttributeError):
            if isinstance(expr, dict):
                k = [sub_expr(sympify(a), dummies_dict) for a in expr]
                v = [sub_expr(sympify(a), dummies_dict) for a in expr.values()]
                expr = dict(zip(k, v))
            elif isinstance(expr, tuple):
                expr = tuple(sub_expr(sympify(a), dummies_dict) for a in expr)
            elif isinstance(expr, list):
                expr = [sub_expr(sympify(a), dummies_dict) for a in expr]
        return expr

    # Transform args
    def isiter(l):
        return iterable(l, exclude=(str,))

    if isiter(args) and any(isiter(i) for i in args):
        import re
        dum_args = [lambdarepr(Dummy(str(i))) for i in range(len(args))]
        iter_args = ','.join([i if isiter(a) else i
                              for i, a in zip(dum_args, args)])
        lstr = lambdastr(flatten(args), expr, printer=printer, dummify=dummify)
        flat = '__flatten_args__'
        rv = f"lambda {','.join(dum_args)}: ({lstr})(*list({flat}([{iter_args}])))"
        if len(re.findall(f'\\b{flat}\\b', rv)) > 1:
            raise ValueError(f'the name {flat} is reserved by lambdastr')
        return rv

    dummies_dict = {}
    if dummify:
        args = sub_args(args, dummies_dict)
    else:
        if iterable(args):
            args = ','.join(str(a) for a in args)

    # Transform expr
    if dummify:
        if isinstance(expr, str):
            pass
        else:
            expr = sub_expr(expr, dummies_dict)
    expr = lambdarepr(expr)

    return f'lambda {args}: ({expr})'


def _imp_namespace(expr, namespace=None):
    """Return namespace dict with function implementations

    We need to search for functions in anything that can be thrown at
    us - that is - anything that could be passed as `expr`.  Examples
    include diofant expressions, as well as tuples, lists and dicts that may
    contain diofant expressions.

    Parameters
    ==========

    expr : object
       Something passed to lambdify, that will generate valid code from
       ``str(expr)``.
    namespace : None or mapping
       Namespace to fill.  None results in new empty dict

    Returns
    =======

    namespace : dict
       dict with keys of implemented function names within `expr` and
       corresponding values being the numerical implementation of
       function

    Examples
    ========

    >>> f = implemented_function(Function('f'), lambda x: x+1)
    >>> g = implemented_function(Function('g'), lambda x: x*10)
    >>> namespace = _imp_namespace(f(g(x)))
    >>> sorted(namespace)
    ['f', 'g']

    """
    # Delayed import to avoid circular imports
    from ..core.function import FunctionClass
    if namespace is None:
        namespace = {}
    # tuples, lists, dicts are valid expressions
    if is_sequence(expr):
        for arg in expr:
            _imp_namespace(arg, namespace)
        return namespace
    elif isinstance(expr, dict):
        for key, val in expr.items():
            # functions can be in dictionary keys
            _imp_namespace(key, namespace)
            _imp_namespace(val, namespace)
        return namespace
    # diofant expressions may be Functions themselves
    func = getattr(expr, 'func', None)
    if isinstance(func, FunctionClass):
        imp = getattr(func, '_imp_', None)
        if imp is not None:
            name = expr.func.__name__
            if name in namespace and namespace[name] != imp:
                raise ValueError(f'We found more than one implementation with name "{name}"')
            namespace[name] = imp
    # and / or they may take Functions as arguments
    if hasattr(expr, 'args'):
        for arg in expr.args:
            _imp_namespace(arg, namespace)
    return namespace


def implemented_function(symfunc, implementation):
    """Add numerical ``implementation`` to function ``symfunc``.

    ``symfunc`` can be an ``UndefinedFunction`` instance, or a name string.
    In the latter case we create an ``UndefinedFunction`` instance with that
    name.

    Be aware that this is a quick workaround, not a general method to create
    special symbolic functions. If you want to create a symbolic function to be
    used by all the machinery of Diofant you should subclass the ``Function``
    class.

    Parameters
    ==========

    symfunc : ``str`` or ``UndefinedFunction`` instance
       If ``str``, then create new ``UndefinedFunction`` with this as
       name.  If `symfunc` is a diofant function, attach implementation to it.
    implementation : callable
       numerical implementation to be called by ``evalf()`` or ``lambdify``

    Returns
    =======

    afunc : diofant.FunctionClass instance
       function with attached implementation

    Examples
    ========

    >>> f = implemented_function(Function('f'), lambda x: x+1)
    >>> lam_f = lambdify(x, f(x))
    >>> lam_f(4)
    5

    """
    # Delayed import to avoid circular imports
    from ..core.function import UndefinedFunction

    # if name, create function to hold implementation
    if isinstance(symfunc, str):
        symfunc = UndefinedFunction(symfunc)
    elif not isinstance(symfunc, UndefinedFunction):
        raise ValueError('symfunc should be either a string or'
                         ' an UndefinedFunction instance.')
    # We need to attach as a method because symfunc will be a class
    symfunc._imp_ = staticmethod(implementation)
    return symfunc



"""Miscellaneous stuff that doesn't really fit anywhere else."""

import textwrap


def filldedent(s, w=70):
    """
    Strips leading and trailing empty lines from a copy of `s`, then dedents,
    fills and returns it.

    Empty line stripping serves to deal with docstrings like this one that
    start with a newline after the initial triple quote, inserting an empty
    line at the beginning of the string.
    """
    return '\n' + textwrap.fill(textwrap.dedent(str(s)).strip('\n'), width=w)



import operator
import random

from ..core import Dummy
from ..ntheory import nextprime
from ..ntheory.modular import crt, integer_rational_reconstruction
from . import rings
from .orderings import build_product_order
from .polyerrors import ModularGCDFailed


def _chinese_remainder_reconstruction(hp, hq, p, q):
    r"""
    Construct a polynomial `h_{pq}` in
    `\mathbb{Z}_{p q}[x_0, \ldots, x_{k-1}]` such that

    .. math ::

        h_{pq} = h_p \; \mathrm{mod} \, p

        h_{pq} = h_q \; \mathrm{mod} \, q

    for relatively prime integers `p` and `q` and polynomials
    `h_p` and `h_q` in `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]` and
    `\mathbb{Z}_q[x_0, \ldots, x_{k-1}]` respectively.

    The coefficients of the polynomial `h_{pq}` are computed with the
    Chinese Remainder Theorem. The symmetric representation in
    `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]`,
    `\mathbb{Z}_q[x_0, \ldots, x_{k-1}]` and
    `\mathbb{Z}_{p q}[x_0, \ldots, x_{k-1}]` is used.

    Parameters
    ==========

    hp : PolyElement
        multivariate integer polynomial with coefficients in `\mathbb{Z}_p`
    hq : PolyElement
        multivariate integer polynomial with coefficients in `\mathbb{Z}_q`
    p : Integer
        modulus of `h_p`, relatively prime to `q`
    q : Integer
        modulus of `h_q`, relatively prime to `p`

    Examples
    ========

    >>> _, x, y = ring('x y', ZZ)
    >>> p = 3
    >>> q = 5

    >>> hp = x**3*y - x**2 - 1
    >>> hq = -x**3*y - 2*x*y**2 + 2

    >>> hpq = _chinese_remainder_reconstruction(hp, hq, p, q)
    >>> hpq
    4*x**3*y + 5*x**2 + 3*x*y**2 + 2

    >>> hpq.trunc_ground(p) == hp
    True
    >>> hpq.trunc_ground(q) == hq
    True

    >>> _, x, y, z = ring('x y z', ZZ)
    >>> p = 6
    >>> q = 5

    >>> hp = 3*x**4 - y**3*z + z
    >>> hq = -2*x**4 + z

    >>> hpq = _chinese_remainder_reconstruction(hp, hq, p, q)
    >>> hpq
    3*x**4 + 5*y**3*z + z

    >>> hpq.trunc_ground(p) == hp
    True
    >>> hpq.trunc_ground(q) == hq
    True

    """
    hpmonoms = set(hp)
    hqmonoms = set(hq)
    monoms = hpmonoms & hqmonoms
    hpmonoms -= monoms
    hqmonoms -= monoms

    zero = hp.ring.domain.zero

    hpq = hp.ring.zero

    if isinstance(hp.ring.domain, rings.PolynomialRing):
        crt_ = _chinese_remainder_reconstruction
    else:
        def crt_(cp, cq, p, q):
            return crt([p, q], [cp, cq], symmetric=True)[0]

    for monom in monoms:
        hpq[monom] = crt_(hp[monom], hq[monom], p, q)
    for monom in hpmonoms:
        hpq[monom] = crt_(hp[monom], zero, p, q)
    for monom in hqmonoms:
        hpq[monom] = crt_(zero, hq[monom], p, q)

    return hpq


def _interpolate(evalpoints, hpeval, ring, i, p, ground=False):
    r"""
    Reconstruct a polynomial `h_p` in `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]`
    from a list of evaluation points in `\mathbb{Z}_p` and a list of
    polynomials in
    `\mathbb{Z}_p[x_0, \ldots, x_{i-1}, x_{i+1}, \ldots, x_{k-1}]`, which
    are the images of `h_p` evaluated in the variable `x_i`.

    It is also possible to reconstruct a parameter of the ground domain,
    i.e. if `h_p` is a polynomial over `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]`.
    In this case, one has to set ``ground=True``.

    Parameters
    ==========

    evalpoints : list of Integer objects
        list of evaluation points in `\mathbb{Z}_p`
    hpeval : list of PolyElement objects
        list of polynomials in (resp. over)
        `\mathbb{Z}_p[x_0, \ldots, x_{i-1}, x_{i+1}, \ldots, x_{k-1}]`,
        images of `h_p` evaluated in the variable `x_i`
    ring : PolynomialRing
        `h_p` will be an element of this ring
    i : Integer
        index of the variable which has to be reconstructed
    p : Integer
        prime number, modulus of `h_p`
    ground : Boolean
        indicates whether `x_i` is in the ground domain, default is
        ``False``

    Returns
    =======

    hp : PolyElement
        interpolated polynomial in (resp. over)
        `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]`

    """
    hp = ring.zero

    if ground:
        domain = ring.domain.domain
        y = ring.domain.gens[i]
    else:
        domain = ring.domain
        y = ring.gens[i]

    for a, hpa in zip(evalpoints, hpeval):
        numer = ring.one
        denom = domain.one
        for b in evalpoints:
            if b == a:
                continue

            numer *= y - b
            denom *= a - b

        if ground:
            denom = domain.invert(denom, p)
        else:
            denom = denom**-1

        coeff = numer*denom
        hp += hpa * coeff

    return hp


def _modgcd_p(f, g, degbound, contbound):
    r"""
    Compute the GCD of two polynomials in `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]`.

    The algorithm reduces the problem step by step by evaluating the
    polynomials `f` and `g` at `x_{k-1} = a` for suitable
    `a \in \mathbb{Z}_p` and then calls itself recursively to compute the GCD
    in `\mathbb{Z}_p[x_0, \ldots, x_{k-2}]`. If these recursive calls are
    successful for enough evaluation points, the GCD in `k` variables is
    interpolated, otherwise the algorithm returns ``None``. Every time a GCD
    or a content is computed, their degrees are compared with the bounds. If
    a degree greater then the bound is encountered, then the current call
    returns ``None`` and a new evaluation point has to be chosen. If at some
    point the degree is smaller, the correspondent bound is updated and the
    algorithm fails.

    Parameters
    ==========

    f : PolyElement
        multivariate polynomial with coefficients in `\mathbb{Z}_p`
    g : PolyElement
        multivariate polynomial with coefficients in `\mathbb{Z}_p`
    degbound : list of Integer objects
        ``degbound[i]`` is an upper bound for the degree of the GCD of `f`
        and `g` in the variable `x_i`
    contbound : list of Integer objects
        ``contbound[i]`` is an upper bound for the degree of the content of
        the GCD in `\mathbb{Z}_p[x_i][x_0, \ldots, x_{i-1}]`, ``contbound[0]``
        is not used can therefore be chosen arbitrarily.

    Returns
    =======

    h : PolyElement
        GCD of the polynomials `f` and `g` or ``None``

    References
    ==========

    * :cite:`Monagan2000Brown`
    * :cite:`Brown1971gcd`

    """
    ring = f.ring
    domain = ring.domain
    k = ring.ngens

    if ring.is_univariate:
        h = ring.gcd(f, g)

        if (degh := h.degree()) > degbound[0]:
            return
        elif degh < degbound[0]:
            degbound[0] = degh
            raise ModularGCDFailed

        return h

    yring = ring.eject(-1).domain  # Z_p[y]

    degyf = f.degree(-1)
    degyg = g.degree(-1)

    contf, f = f.eject(-1).primitive()
    contg, g = g.eject(-1).primitive()

    yf, yg = f, g

    f = f.inject()
    g = g.inject()

    conth = yring.gcd(contf, contg)
    conth = conth.set_ring(ring)

    if (degconth := conth.degree(-1)) > contbound[k-1]:
        return
    elif degconth < contbound[k-1]:
        contbound[k-1] = degconth
        raise ModularGCDFailed

    delta = yring.gcd(yf.LC, yg.LC)

    evaltest = delta

    for i in range(k-1):
        order = build_product_order((('lex', i),
                                     ('lex', *range(k-1))),
                                    list(range(k-1)))
        evaltest *= yring.gcd(yf[yf.leading_expv(order)],
                              yg[yg.leading_expv(order)])

    N = min(degyf - contf.degree(), degyg - contg.degree(),
            degbound[k-1] - contbound[k-1] + delta.degree()) + 1

    p = domain.characteristic
    if p < N:
        return

    n = 0
    d = 0
    evalpoints = []
    heval = []
    points = list(map(domain, range(p)))
    random.shuffle(points)

    while points:
        a = points.pop()

        if not evaltest.eval(0, a):
            continue

        fa = f.eval(-1, a)
        ga = g.eval(-1, a)

        # polynomials in Z_p[x_0, ..., x_{k-2}]
        ha = _modgcd_p(fa, ga, degbound, contbound)

        if ha is None:
            if d < n:
                d += 1
                continue
            return
        else:
            n += 1

        if ha.is_ground:
            return conth

        ha *= delta.eval(0, a)

        evalpoints.append(a)
        heval.append(ha)

        if n == N:
            h = _interpolate(evalpoints, heval, ring, -1, p)
            h = h.eject(-1).primitive()[1].inject()*conth

            if (degyh := h.degree(-1)) < degbound[k-1]:
                degbound[k-1] = degyh
                raise ModularGCDFailed

            return h


def modgcd(f, g):
    r"""
    Compute the GCD of two polynomials in `\mathbb{Z}[x_0, \ldots, x_{k-1}]`
    using a modular algorithm.

    The algorithm computes the GCD of two multivariate integer polynomials
    `f` and `g` by calculating the GCD in
    `\mathbb{Z}_p[x_0, \ldots, x_{k-1}]` for suitable primes `p` and then
    reconstructing the coefficients with the Chinese Remainder Theorem. To
    compute the multivariate GCD over `\mathbb{Z}_p` the recursive
    subroutine ``_modgcd_p`` is used. To verify the result in
    `\mathbb{Z}[x_0, \ldots, x_{k-1}]`, trial division is done, but only for
    candidates which are very likely the desired GCD.

    Parameters
    ==========

    f : PolyElement
        multivariate integer polynomial
    g : PolyElement
        multivariate integer polynomial

    Returns
    =======

    h : PolyElement
        GCD of the polynomials `f` and `g`

    Examples
    ========

    >>> _, x, y = ring('x y', ZZ)

    >>> modgcd((x - y)*(x + y), (x + y)**2)
    x + y

    >>> _, x, y, z = ring('x y z', ZZ)

    >>> modgcd((x - y)*z**2, (x**2 + 1)*z)
    z

    References
    ==========

    * :cite:`Monagan2000Brown`
    * :cite:`Brown1971gcd`

    """
    assert f.ring == g.ring
    assert f.ring.domain.is_IntegerRing

    ring = f.ring
    domain = ring.domain
    k = ring.ngens

    # divide out integer content
    cf, f = f.primitive()
    cg, g = g.primitive()
    ch = ring.domain.gcd(cf, cg)

    gamma = ring.domain.gcd(f.LC, g.LC)

    badprimes = ring.domain.one
    for i in range(k):
        order = build_product_order((('lex', i),
                                     ('lex', *range(k))),
                                    list(range(k)))
        badprimes *= ring.domain.gcd(f[f.leading_expv(order)],
                                     g[g.leading_expv(order)])

    degbound = [min(f.degree(x), g.degree(x)) for x in ring.gens]
    contbound = list(degbound)

    m = 1
    p = 1

    while True:
        p = nextprime(p)
        while badprimes % p == 0:
            p = nextprime(p)

        fp, gp = map(operator.methodcaller('set_domain',
                                           domain.finite_field(p)), (f, g))

        try:
            # monic GCD of fp, gp in Z_p[x_0, ..., x_{k-2}, y]
            hp = _modgcd_p(fp, gp, degbound, contbound)
        except ModularGCDFailed:
            m = 1
            continue

        if hp is None:
            continue

        hp *= gamma
        if m == 1:
            m = p
            hlastm = hp
            continue

        hm = _chinese_remainder_reconstruction(hp, hlastm, p, m)
        m *= p
        hm = hm.set_domain(domain)

        if not hm == hlastm:
            hlastm = hm
            continue

        h = hm.primitive()[1]
        if not f % h and not g % h:
            return h*ch


def _rational_function_reconstruction(c, p, m):
    r"""
    Reconstruct a rational function `\frac a b` in `\mathbb Z_p(t)` from

    .. math::

        c = \frac a b \; \mathrm{mod} \, m,

    where `c` and `m` are polynomials in `\mathbb Z_p[t]` and `m` has
    positive degree.

    The algorithm is based on the Euclidean Algorithm. In general, `m` is
    not irreducible, so it is possible that `b` is not invertible modulo
    `m`. In that case ``None`` is returned.

    Parameters
    ==========

    c : PolyElement
        univariate polynomial in `\mathbb Z[t]`
    p : Integer
        prime number
    m : PolyElement
        modulus, not necessarily irreducible

    Returns
    =======

    frac : FracElement
        either `\frac a b` in `\mathbb Z(t)` or ``None``

    References
    ==========

    * :cite:`Monagan2004algebraic`

    """
    ring = c.ring
    field = ring.field
    domain = ring.domain
    M = m.degree()
    N = M // 2
    D = M - N - 1

    pdomain = domain.finite_field(p)
    pring = ring.clone(domain=pdomain)
    c, m = map(operator.methodcaller('set_domain', pdomain), (c, m))

    r0, s0 = m, pring.zero
    r1, s1 = c, pring.one

    while r1.degree() > N:
        quo = r0 // r1
        r0, r1 = r1, r0 - quo*r1
        s0, s1 = s1, s0 - quo*s1

    if s1.degree() > D or pring.gcd(s1, m) != 1:
        return

    lc = s1.LC
    if lc != 1:
        r1 /= lc
        s1 /= lc

    r1, s1 = map(operator.methodcaller('set_domain', domain), (r1, s1))

    return field(r1) / field(s1)


def _rational_reconstruction_func_coeffs(hm, p, m, ring, k):
    r"""
    Reconstruct every coefficient `c_h` of a polynomial `h` in
    `\mathbb Z_p(t_k)[t_1, \ldots, t_{k-1}][x, z]` from the corresponding
    coefficient `c_{h_m}` of a polynomial `h_m` in
    `\mathbb Z_p[t_1, \ldots, t_k][x, z] \cong \mathbb Z_p[t_k][t_1, \ldots, t_{k-1}][x, z]`
    such that

    .. math::

        c_{h_m} = c_h \; \mathrm{mod} \, m,

    where `m \in \mathbb Z_p[t]`.

    The reconstruction is based on the Euclidean Algorithm. In general, `m`
    is not irreducible, so it is possible that this fails for some
    coefficient. In that case ``None`` is returned.

    Parameters
    ==========

    hm : PolyElement
        polynomial in `\mathbb Z[t_1, \ldots, t_k][x, z]`
    p : Integer
        prime number, modulus of `\mathbb Z_p`
    m : PolyElement
        modulus, polynomial in `\mathbb Z[t]`, not necessarily irreducible
    ring : PolynomialRing
        `\mathbb Z(t_k)[t_1, \ldots, t_{k-1}][x, z]`, `h` will be an
        element of this ring
    k : Integer
        index of the parameter `t_k` which will be reconstructed

    Returns
    =======

    h : PolyElement
        reconstructed polynomial in
        `\mathbb Z(t_k)[t_1, \ldots, t_{k-1}][x, z]` or ``None``

    See also
    ========

    _rational_function_reconstruction

    """
    h = ring.zero

    for monom, coeff in hm.items():
        if k == 0:
            coeffh = _rational_function_reconstruction(coeff, p, m)

            if not coeffh:
                return

        else:
            coeffh = ring.domain.zero
            for mon, c in coeff.eject(k).items():
                ch = _rational_function_reconstruction(c, p, m)

                if not ch:
                    return

                coeffh[mon] = ch

        h[monom] = coeffh

    return h


def _gf_gcdex(f, g, p):
    r"""
    Extended Euclidean Algorithm for two univariate polynomials over
    `\mathbb Z_p`.

    Returns polynomials `s, t` and `h`, such that `h` is the GCD of `f` and
    `g` and `sf + tg = h \; \mathrm{mod} \, p`.

    """
    ring = f.ring
    dom = ring.domain
    domp = dom.finite_field(p)
    f, g = map(operator.methodcaller('set_domain', domp), (f, g))
    return tuple(map(operator.methodcaller('set_domain', dom), f.gcdex(g)))


def _trunc(f, minpoly, p):
    r"""
    Compute the reduced representation of a polynomial `f` in
    `\mathbb Z_p[z] / (\check m_{\alpha}(z))[x]`

    Parameters
    ==========

    f : PolyElement
        polynomial in `\mathbb Z[x, z]`
    minpoly : PolyElement
        polynomial `\check m_{\alpha} \in \mathbb Z[z]`, not necessarily
        irreducible
    p : Integer
        prime number, modulus of `\mathbb Z_p`

    Returns
    =======

    ftrunc : PolyElement
        polynomial in `\mathbb Z[x, z]`, reduced modulo
        `\check m_{\alpha}(z)` and `p`

    """
    ring = f.ring
    minpoly = minpoly.set_ring(ring)
    p_ = ring.ground_new(p)

    return f.trunc_ground(p).div([minpoly, p_])[1].trunc_ground(p)


def _euclidean_algorithm(f, g, minpoly, p):
    r"""
    Compute the monic GCD of two univariate polynomials in
    `\mathbb{Z}_p[z]/(\check m_{\alpha}(z))[x]` with the Euclidean
    Algorithm.

    In general, `\check m_{\alpha}(z)` is not irreducible, so it is possible
    that some leading coefficient is not invertible modulo
    `\check m_{\alpha}(z)`. In that case ``None`` is returned.

    Parameters
    ==========

    f, g : PolyElement
        polynomials in `\mathbb Z[x, z]`
    minpoly : PolyElement
        polynomial in `\mathbb Z[z]`, not necessarily irreducible
    p : Integer
        prime number, modulus of `\mathbb Z_p`

    Returns
    =======

    h : PolyElement
        GCD of `f` and `g` in `\mathbb Z[z, x]` or ``None``, coefficients
        are in `\left[ -\frac{p-1} 2, \frac{p-1} 2 \right]`

    """
    ring = f.ring
    domain = ring.domain

    f = _trunc(f, minpoly, p)
    g = _trunc(g, minpoly, p)

    while g:
        rem = f
        deg = g.degree(0)  # degree in x
        lcinv, _, gcd = _gf_gcdex(g.eject(-1).LC, minpoly, p)

        if not gcd == 1:
            return

        while True:
            degrem = rem.degree(0)  # degree in x
            if degrem < deg:
                break
            quo = (lcinv * rem.eject(-1).LC).set_ring(ring)
            m = ring.from_terms([((degrem - deg, 0), domain.one)])
            rem = _trunc(rem - g*m*quo, minpoly, p)

        f = g
        g = rem

    lcfinv = _gf_gcdex(f.eject(-1).LC, minpoly, p)[0].set_ring(ring)

    return _trunc(f * lcfinv, minpoly, p)


def trial_division(f, h, minpoly, p=None):
    r"""
    Check if `h` divides `f` in
    `\mathbb K[t_1, \ldots, t_k][z]/(m_{\alpha}(z))`, where `\mathbb K` is
    either `\mathbb Q` or `\mathbb Z_p`.

    This algorithm is based on pseudo division and does not use any
    fractions. By default `\mathbb K` is `\mathbb Q`, if a prime number `p`
    is given, `\mathbb Z_p` is chosen instead.

    Parameters
    ==========

    f, h : PolyElement
        polynomials in `\mathbb Z[t_1, \ldots, t_k][x, z]`
    minpoly : PolyElement
        polynomial `m_{\alpha}(z)` in `\mathbb Z[t_1, \ldots, t_k][z]`
    p : Integer or None
        if `p` is given, `\mathbb K` is set to `\mathbb Z_p` instead of
        `\mathbb Q`, default is ``None``

    Returns
    =======

    rem : PolyElement
        remainder of `\frac f h`

    References
    ==========

    * :cite:`vanHoeij2002modgcd`

    """
    ring = f.ring
    domain = ring.domain
    zxring = ring.clone(symbols=(ring.symbols[1], ring.symbols[0]))
    minpoly = minpoly.set_ring(ring)
    rem = f

    degrem = rem.degree()
    degh = h.degree()
    degm = minpoly.degree(1)

    lch = h.eject(-1).LC.set_ring(ring)
    lcm = minpoly.LC

    while rem and degrem >= degh:
        # polynomial in Z[t_1, ..., t_k][z]
        lcrem = rem.eject(-1).LC.set_ring(ring)
        rem = rem*lch - h*ring.from_terms([((degrem - degh, 0),
                                            domain.one)])*lcrem
        if p:
            rem = rem.trunc_ground(p)
        degrem = rem.degree(1)

        while rem and degrem >= degm:
            # polynomial in Z[t_1, ..., t_k][x]
            lcrem = rem.set_ring(zxring).eject(-1).LC.set_ring(ring)
            rem = rem*lcm - minpoly*ring.from_terms([((0, degrem - degm),
                                                     domain.one)])*lcrem
            if p:
                rem = rem.trunc_ground(p)
            degrem = rem.degree(1)

        degrem = rem.degree()

    return rem


def _evaluate_ground(f, i, a):
    r"""
    Evaluate a polynomial `f` at `a` in the `i`-th variable of the ground
    domain.

    """
    ring = f.ring.clone(domain=f.ring.domain.ring.drop(i))
    fa = ring.zero

    for monom, coeff in f.items():
        fa[monom] = coeff.eval(i, a)

    return fa


def _func_field_modgcd_p(f, g, minpoly, p):
    r"""
    Compute the GCD of two polynomials `f` and `g` in
    `\mathbb Z_p(t_1, \ldots, t_k)[z]/(\check m_\alpha(z))[x]`.

    The algorithm reduces the problem step by step by evaluating the
    polynomials `f` and `g` at `t_k = a` for suitable `a \in \mathbb Z_p`
    and then calls itself recursively to compute the GCD in
    `\mathbb Z_p(t_1, \ldots, t_{k-1})[z]/(\check m_\alpha(z))[x]`. If these
    recursive calls are successful, the GCD over `k` variables is
    interpolated, otherwise the algorithm returns ``None``. After
    interpolation, Rational Function Reconstruction is used to obtain the
    correct coefficients. If this fails, a new evaluation point has to be
    chosen, otherwise the desired polynomial is obtained by clearing
    denominators. The result is verified with a fraction free trial
    division.

    Parameters
    ==========

    f, g : PolyElement
        polynomials in `\mathbb Z[t_1, \ldots, t_k][x, z]`
    minpoly : PolyElement
        polynomial in `\mathbb Z[t_1, \ldots, t_k][z]`, not necessarily
        irreducible
    p : Integer
        prime number, modulus of `\mathbb Z_p`

    Returns
    =======

    h : PolyElement
        primitive associate in `\mathbb Z[t_1, \ldots, t_k][x, z]` of the
        GCD of the polynomials `f` and `g`  or ``None``, coefficients are
        in `\left[ -\frac{p-1} 2, \frac{p-1} 2 \right]`

    References
    ==========

    * :cite:`Monagan2004algebraic`

    """
    ring = f.ring
    domain = ring.domain  # Z[t_1, ..., t_k]

    if isinstance(domain, rings.PolynomialRing):
        k = domain.ngens
    else:
        return _euclidean_algorithm(f, g, minpoly, p)

    if k == 1:
        qdomain = domain.ring.field
    else:
        qdomain = domain.ring.eject(k - 1)
        qdomain = qdomain.clone(domain=qdomain.domain.ring.field)

    qring = ring.clone(domain=qdomain)  # = Z(t_k)[t_1, ..., t_{k-1}][x, z]

    n = 1
    d = 1

    # polynomial in Z_p[t_1, ..., t_k][z]
    gamma = f.eject(-1).LC * g.eject(-1).LC
    # polynomial in Z_p[t_1, ..., t_k]
    delta = minpoly.LC

    evalpoints = []
    heval = []
    LMlist = []
    points = list(range(p))
    random.shuffle(points)

    while points:
        a = points.pop()

        if k == 1:
            test = delta.eval(k-1, a) % p == 0
        else:
            test = delta.eval(k-1, a).trunc_ground(p) == 0

        if test:
            continue

        gammaa = _evaluate_ground(gamma, k-1, a)
        minpolya = _evaluate_ground(minpoly, k-1, a)

        if gammaa.div([minpolya, gammaa.ring(p)])[1] == 0:
            continue

        fa = _evaluate_ground(f, k-1, a)
        ga = _evaluate_ground(g, k-1, a)

        # polynomial in Z_p[x, t_1, ..., t_{k-1}, z]/(minpoly)
        ha = _func_field_modgcd_p(fa, ga, minpolya, p)

        if ha is None:
            if d < n:
                d += 1
                continue
            return

        if ha == 1:
            return ha

        LM = [ha.degree()] + [0]*(k-1)
        if k > 1:
            for monom, coeff in ha.items():
                if monom[0] == LM[0] and coeff.LM > tuple(LM[1:]):
                    LM[1:] = coeff.LM

        evalpoints_a = [a]
        heval_a = [ha]
        if k == 1:
            m = qring.domain.ring.one
        else:
            m = qring.domain.domain.ring.one

        t = m.ring.gens[0]

        for b, hb, LMhb in zip(evalpoints, heval, LMlist):
            if LMhb == LM:
                evalpoints_a.append(b)
                heval_a.append(hb)
                m *= (t - b)

        m = m.trunc_ground(p)
        evalpoints.append(a)
        heval.append(ha)
        LMlist.append(LM)
        n += 1

        # polynomial in Z_p[t_1, ..., t_k][x, z]
        h = _interpolate(evalpoints_a, heval_a, ring, k-1, p, ground=True)

        # polynomial in Z_p(t_k)[t_1, ..., t_{k-1}][x, z]
        h = _rational_reconstruction_func_coeffs(h, p, m, qring, k-1)

        if h is None:
            continue

        if k == 1:
            dom = qring.domain.field
            domp = dom.ring.domain.finite_field(p)
            new_dom = dom.clone(domain=domp)
            h = h.set_domain(new_dom)
            _, h = h.clear_denoms()
            h = h.set_domain(domain)
        else:
            dom = qring.domain.domain.field
            domp = dom.ring.domain.finite_field(p)
            new_dom = dom.clone(domain=domp)
            den = new_dom.ring.one

            for c in h.values():
                c = c.set_domain(new_dom)
                c, _ = c.clear_denoms()
                den = den.ring.lcm(den, c)

            den = new_dom((den, new_dom.ring.one))

            for key, coeff in h.items():
                coeff = coeff.set_domain(new_dom)
                h[key] = (coeff*den).set_domain(dom)

            h = ring.from_expr(h.ring.to_expr(h))

        if not trial_division(f, h, minpoly, p) and not trial_division(g, h, minpoly, p):
            return h


def _rational_reconstruction_int_coeffs(hm, m, ring):
    r"""
    Reconstruct every rational coefficient `c_h` of a polynomial `h` in
    `\mathbb Q[t_1, \ldots, t_k][x, z]` from the corresponding integer
    coefficient `c_{h_m}` of a polynomial `h_m` in
    `\mathbb Z[t_1, \ldots, t_k][x, z]` such that

    .. math::

        c_{h_m} = c_h \; \mathrm{mod} \, m,

    where `m \in \mathbb Z`.

    The reconstruction is based on the Euclidean Algorithm. In general,
    `m` is not a prime number, so it is possible that this fails for some
    coefficient. In that case ``None`` is returned.

    Parameters
    ==========

    hm : PolyElement
        polynomial in `\mathbb Z[t_1, \ldots, t_k][x, z]`
    m : Integer
        modulus, not necessarily prime
    ring : PolynomialRing
        `\mathbb Q[t_1, \ldots, t_k][x, z]`, `h` will be an element of this
        ring

    Returns
    =======

    h : PolyElement
        reconstructed polynomial in `\mathbb Q[t_1, \ldots, t_k][x, z]` or
        ``None``

    See also
    ========

    diofant.ntheory.modular.integer_rational_reconstruction

    """
    h = ring.zero

    if isinstance(ring.domain, rings.PolynomialRing):
        reconstruction = _rational_reconstruction_int_coeffs
        domain = ring.domain.ring
    else:
        reconstruction = integer_rational_reconstruction
        domain = hm.ring.domain

    for monom, coeff in hm.items():
        coeffh = reconstruction(coeff, m, domain)

        if not coeffh:
            return

        h[monom] = coeffh

    return h


def _func_field_modgcd_m(f, g, minpoly):
    r"""
    Compute the GCD of two polynomials in
    `\mathbb Q(t_1, \ldots, t_k)[z]/(m_{\alpha}(z))[x]` using a modular
    algorithm.

    The algorithm computes the GCD of two polynomials `f` and `g` by
    calculating the GCD in
    `\mathbb Z_p(t_1, \ldots, t_k)[z] / (\check m_{\alpha}(z))[x]` for
    suitable primes `p` and the primitive associate `\check m_{\alpha}(z)`
    of `m_{\alpha}(z)`. Then the coefficients are reconstructed with the
    Chinese Remainder Theorem and Rational Reconstruction. To compute the
    GCD over `\mathbb Z_p(t_1, \ldots, t_k)[z] / (\check m_{\alpha})[x]`,
    the recursive subroutine ``_func_field_modgcd_p`` is used. To verify the
    result in `\mathbb Q(t_1, \ldots, t_k)[z] / (m_{\alpha}(z))[x]`, a
    fraction free trial division is used.

    Parameters
    ==========

    f, g : PolyElement
        polynomials in `\mathbb Z[t_1, \ldots, t_k][x, z]`
    minpoly : PolyElement
        irreducible polynomial in `\mathbb Z[t_1, \ldots, t_k][z]`

    Returns
    =======

    h : PolyElement
        the primitive associate in `\mathbb Z[t_1, \ldots, t_k][x, z]` of
        the GCD of `f` and `g`

    Examples
    ========

    >>> _, x, z = ring('x z', ZZ)
    >>> minpoly = (z**2 - 2).drop(0)

    >>> f = x**2 + 2*x*z + 2
    >>> g = x + z
    >>> _func_field_modgcd_m(f, g, minpoly)
    x + z

    >>> D, t = ring('t', ZZ)
    >>> _, x, z = ring('x z', D)
    >>> minpoly = (z**2-3).drop(0)

    >>> f = x**2 + (t + 1)*x*z + 3*t
    >>> g = x*z + 3*t
    >>> _func_field_modgcd_m(f, g, minpoly)
    x + t*z

    References
    ==========

    * :cite:`Monagan2004algebraic`

    See also
    ========

    _func_field_modgcd_p

    """
    ring = f.ring
    domain = ring.domain

    if isinstance(domain, rings.PolynomialRing):
        k = domain.ngens
        QQdomain = domain.ring.clone(domain=domain.domain.field)
        QQring = ring.clone(domain=QQdomain)
    else:
        k = 0
        QQring = ring.clone(domain=domain.field)

    cf, f = f.primitive()
    cg, g = g.primitive()

    # polynomial in Z[t_1, ..., t_k][z]
    gamma = f.eject(-1).LC * g.eject(-1).LC
    # polynomial in Z[t_1, ..., t_k]
    delta = minpoly.LC
    assert k > 0 or delta == 1

    p = 1
    primes = []
    hplist = []
    LMlist = []

    while True:
        p = nextprime(p)

        if gamma.trunc_ground(p) == 0:
            continue

        fp = f.trunc_ground(p)
        gp = g.trunc_ground(p)
        minpolyp = minpoly.trunc_ground(p)

        hp = _func_field_modgcd_p(fp, gp, minpolyp, p)

        if hp is None:
            continue

        if hp == 1:
            return ring.one

        LM = [hp.degree()] + [0]*k
        if k > 0:
            for monom, coeff in hp.items():
                if monom[0] == LM[0] and coeff.LM > tuple(LM[1:]):
                    LM[1:] = coeff.LM

        hm = hp
        m = p

        for q, hq, LMhq in zip(primes, hplist, LMlist):
            if LMhq == LM:
                hm = _chinese_remainder_reconstruction(hq, hm, q, m)
                m *= q

        primes.append(p)
        hplist.append(hp)
        LMlist.append(LM)

        hm = _rational_reconstruction_int_coeffs(hm, m, QQring)

        if hm is None:
            continue

        if k == 0:
            h = hm.clear_denoms()[1]
        else:
            den = domain.domain.one
            for coeff in hm.values():
                den = domain.domain.lcm(den, coeff.clear_denoms()[0])
            h = hm*den

        # convert back to Z[t_1, ..., t_k][x, z] from Q[t_1, ..., t_k][x, z]
        h = h.set_ring(ring)
        h = h.primitive()[1]

        if (not trial_division(f*cf, h, minpoly) and
                not trial_division(g*cg, h, minpoly)):
            return h


def _to_ZZ_poly(f, ring):
    r"""
    Compute an associate of a polynomial
    `f \in \mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]` in
    `\mathbb Z[x_1, \ldots, x_{n-1}][z] / (\check m_{\alpha}(z))[x_0]`,
    where `\check m_{\alpha}(z) \in \mathbb Z[z]` is the primitive associate
    of the minimal polynomial `m_{\alpha}(z)` of `\alpha` over
    `\mathbb Q`.

    Parameters
    ==========

    f : PolyElement
        polynomial in `\mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]`
    ring : PolynomialRing
        `\mathbb Z[x_1, \ldots, x_{n-1}][x_0, z]`

    Returns
    =======

    f_ : PolyElement
        associate of `f` in
        `\mathbb Z[x_1, \ldots, x_{n-1}][x_0, z]`

    """
    f_ = ring.zero

    if isinstance(ring.domain, rings.PolynomialRing):
        domain = ring.domain.domain
    else:
        domain = ring.domain

    den = domain.one

    for coeff in f.values():
        for c in coeff.rep.all_coeffs():
            if c:
                den = domain.lcm(den, c.denominator)

    for monom, coeff in f.items():
        coeff = coeff.rep.all_coeffs()
        m = ring.domain.one
        if isinstance(ring.domain, rings.PolynomialRing):
            m *= ring.domain.from_terms([(monom[1:], domain.one)])
        n = len(coeff)

        for i in range(n):
            if coeff[n - i - 1]:
                c = domain(coeff[n - i - 1] * den) * m

                if (monom[0], n-i-1) not in f_:
                    f_[(monom[0], n-i-1)] = c
                else:
                    f_[(monom[0], n-i-1)] += c

    return f_


def _to_ANP_poly(f, ring):
    r"""
    Convert a polynomial
    `f \in \mathbb Z[x_1, \ldots, x_{n-1}][z]/(\check m_{\alpha}(z))[x_0]`
    to a polynomial in `\mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]`,
    where `\check m_{\alpha}(z) \in \mathbb Z[z]` is the primitive associate
    of the minimal polynomial `m_{\alpha}(z)` of `\alpha` over
    `\mathbb Q`.

    Parameters
    ==========

    f : PolyElement
        polynomial in `\mathbb Z[x_1, \ldots, x_{n-1}][x_0, z]`
    ring : PolynomialRing
        `\mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]`

    Returns
    =======

    f_ : PolyElement
        polynomial in `\mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]`

    """
    domain = ring.domain
    f_ = ring.zero

    if isinstance(f.ring.domain, rings.PolynomialRing):
        for monom, coeff in f.items():
            for mon, coef in coeff.items():
                m = (monom[0],) + mon
                c = domain([0]*monom[1] + [domain.domain(coef)])

                if m not in f_:
                    f_[m] = c
                else:
                    f_[m] += c

    else:
        for monom, coeff in f.items():
            m = monom[0],
            c = domain([0]*monom[1] + [domain.domain(coeff)])

            if m not in f_:
                f_[m] = c
            else:
                f_[m] += c

    return f_


def _minpoly_from_dense(minpoly, ring):
    r"""
    Change representation of the minimal polynomial from ``Poly`` to
    ``PolyElement`` for a given ring.

    """
    minpoly_ = ring.zero

    for monom, coeff in minpoly.items():
        minpoly_[monom] = ring.domain(coeff)

    return minpoly_


def _primitive_in_x0(f):
    r"""
    Compute the content in `x_0` and the primitive part of a polynomial `f`
    in
    `\mathbb Q(\alpha)[x_0, x_1, \ldots, x_{n-1}] \cong \mathbb Q(\alpha)[x_1, \ldots, x_{n-1}][x_0]`.

    """
    fring = f.ring
    f_ = f.eject(*range(1, fring.ngens))
    dom = f_.ring.domain.ring
    cont = dom.zero

    for coeff in f_.values():
        cont = func_field_modgcd(cont, coeff)
        if cont == 1:
            return cont, f

    return cont, f//cont.set_ring(fring)


def func_field_modgcd(f, g):
    r"""
    Compute the GCD of two polynomials `f` and `g` in
    `\mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]` using a modular algorithm.

    The algorithm first computes the primitive associate
    `\check m_{\alpha}(z)` of the minimal polynomial `m_{\alpha}` in
    `\mathbb{Z}[z]` and the primitive associates of `f` and `g` in
    `\mathbb{Z}[x_1, \ldots, x_{n-1}][z]/(\check m_{\alpha})[x_0]`. Then it
    computes the GCD in
    `\mathbb Q(x_1, \ldots, x_{n-1})[z]/(m_{\alpha}(z))[x_0]`.
    This is done by calculating the GCD in
    `\mathbb{Z}_p(x_1, \ldots, x_{n-1})[z]/(\check m_{\alpha}(z))[x_0]` for
    suitable primes `p` and then reconstructing the coefficients with the
    Chinese Remainder Theorem and Rational Reconstuction. The GCD over
    `\mathbb{Z}_p(x_1, \ldots, x_{n-1})[z]/(\check m_{\alpha}(z))[x_0]` is
    computed with a recursive subroutine, which evaluates the polynomials at
    `x_{n-1} = a` for suitable evaluation points `a \in \mathbb Z_p` and
    then calls itself recursively until the ground domain does no longer
    contain any parameters. For
    `\mathbb{Z}_p[z]/(\check m_{\alpha}(z))[x_0]` the Euclidean Algorithm is
    used. The results of those recursive calls are then interpolated and
    Rational Function Reconstruction is used to obtain the correct
    coefficients. The results, both in
    `\mathbb Q(x_1, \ldots, x_{n-1})[z]/(m_{\alpha}(z))[x_0]` and
    `\mathbb{Z}_p(x_1, \ldots, x_{n-1})[z]/(\check m_{\alpha}(z))[x_0]`, are
    verified by a fraction free trial division.

    Apart from the above GCD computation some GCDs in
    `\mathbb Q(\alpha)[x_1, \ldots, x_{n-1}]` have to be calculated,
    because treating the polynomials as univariate ones can result in
    a spurious content of the GCD. For this ``func_field_modgcd`` is
    called recursively.

    Parameters
    ==========

    f, g : PolyElement
        polynomials in `\mathbb Q(\alpha)[x_0, \ldots, x_{n-1}]`

    Returns
    =======

    h : PolyElement
        monic GCD of the polynomials `f` and `g`

    Examples
    ========

    >>> A = QQ.algebraic_field(sqrt(2))
    >>> _, x = ring('x', A)

    >>> func_field_modgcd(x**2 - 2, x + sqrt(2))
    x + sqrt(2)

    >>> _, x, y = ring('x y', A)

    >>> func_field_modgcd((x + sqrt(2)*y)**2, x + sqrt(2)*y)
    x + sqrt(2)*y

    >>> func_field_modgcd(x + sqrt(2)*y, x + y)
    1

    References
    ==========

    * :cite:`Monagan2004algebraic`

    """
    ring = f.ring
    domain = ring.domain
    n = ring.ngens

    assert ring == g.ring
    assert domain.is_AlgebraicField

    if not f:
        return ring._gcd_zero(g)

    z = Dummy('z')

    ZZring = ring.clone(symbols=ring.symbols + (z,), domain=domain.domain.ring)

    if n == 1:
        f_ = _to_ZZ_poly(f, ZZring)
        g_ = _to_ZZ_poly(g, ZZring)
        minpoly = ZZring.drop(0).from_dict(domain.mod)

        h = _func_field_modgcd_m(f_, g_, minpoly)
        h = _to_ANP_poly(h, ring)

    else:
        # contx0f in Q(a)[x_1, ..., x_{n-1}], f in Q(a)[x_0, ..., x_{n-1}]
        contx0f, f = _primitive_in_x0(f)
        contx0g, g = _primitive_in_x0(g)
        contx0h = func_field_modgcd(contx0f, contx0g)

        ZZring_ = ZZring.eject(*range(1, n))

        f_ = _to_ZZ_poly(f, ZZring_)
        g_ = _to_ZZ_poly(g, ZZring_)
        minpoly = _minpoly_from_dense(domain.mod, ZZring_.drop(0))

        h = _func_field_modgcd_m(f_, g_, minpoly)
        h = _to_ANP_poly(h, ring)

        _, h = _primitive_in_x0(h)
        h *= contx0h.set_ring(ring)
        f *= contx0f.set_ring(ring)
        g *= contx0g.set_ring(ring)

    return h.quo_ground(h.LC)



from diofant.utilities.decorator import no_attrs_in_subclass


__all__ = ()


def test_no_attrs_in_subclass():
    class A:
        x = 'test'

    A.x = no_attrs_in_subclass(A, A.x)

    class B(A):
        pass

    assert hasattr(A, 'x') is True
    assert hasattr(B, 'x') is False
